"""
The proper way to create an uncertain array is by calling :func:`.uarray`
"""
# Adding numpy arrays to GTC is not an easy exercise.
# Our need is to provide convenient containers for uncertain numbers.
# We do not try to integrate uncertain numbers in numpy's design.
from __future__ import division

import warnings

from numbers import Number, Real, Complex
from math import isnan, isinf
from cmath import isnan as cisnan
from cmath import isinf as cisinf
try:
    from itertools import izip  # Python 2
except ImportError:
    izip = zip
    xrange = range

import numpy as np

from GTC import is_sequence
from GTC.linear_algebra import matmul

from GTC.core import (
    value,
    uncertainty,
    variance,
    dof,
    cos,
    sin,
    tan,
    acos,
    asin,
    atan,
    atan2,
    exp,
    log,
    log10,
    sqrt,
    sinh,
    cosh,
    tanh,
    acosh,
    asinh,
    atanh,
    mag_squared,
    magnitude,
    phase,
    result,
)

from GTC.lib import (
    UncertainReal,
    UncertainComplex
)


def _isnan(number):
    val = value(number)
    if isinstance(val, Real):
        return isnan(val)
    elif isinstance(val, Complex):
        return cisnan(val)
    else:
        raise TypeError('cannot calculate isnan of type {}'.format(type(number)))


def _isinf(number):
    val = value(number)
    if isinstance(val, Real):
        return isinf(val)
    elif isinstance(val, Complex):
        return cisinf(val)
    else:
        raise TypeError('cannot calculate isinf of type {}'.format(type(number)))

# Note numpy defines its own numeric types, instead of bool, int,
# float, complex, that have additional attributes. These types are needed by
# functions like `numpy.average`. (Uses `dtype` and `.size` attributes
# on the result returned by `mean`, as defined in a subclass if available.)

# One way to fix this is to add the required attributes
# to all the return values from `UncertainArray` methods.

# Another option is to ensure that array elements
# are always numpy-compatible and to ensure that all
# uncertain number objects are initialised with
#           a.dtype = np.dtype('O')
#           a.size = 1
#           a.shape = ()
#
# Our use of `dtype=object` for arrays means that numeric
# elements are not cast to numpy types when loaded into an array.
# To fix this would require iteration through all arrays as they
# are being created!

#--------------------------------------------------------------------
class UncertainArray(np.ndarray):
    """An :class:`UncertainArray` can contain elements of type
    :class:`int`, :class:`float`, :class:`complex`,
    :class:`.UncertainReal` or :class:`.UncertainComplex`.

    Do not instantiate this class directly. Use :func:`~.uarray` instead.

    Base: :class:`numpy.ndarray`

    .. versionadded:: 1.1

    """
    def __new__(cls, array, dtype=None, label=None):
        # The first case allows users to create uarray instances
        # with a definite numpy number type. This could be done
        # by wrapping a call to uarray() around an ndarray.
        # Without this, the type gets converted back to Python.
        if isinstance(array, np.ndarray):
            dtype = array.dtype
        elif dtype is None:
            dtype = np.dtype('O')

        obj = np.asarray(array, dtype=dtype).view(cls)
        obj._label = label
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return

        self._label = getattr(obj, 'label', None)

        # numpy looks at type().__name__ when preparing
        # a string representation of the object. This
        # change means we see `uarray` not `UncertainArray`.
        self.__class__.__name__ = 'uarray'

        self._broadcasted_shape = None

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        try:
            attr = getattr(self, '_' + ufunc.__name__)
        except AttributeError:
            # Want to raise a NotImplementedError without nested exceptions
            # In Python 3 this could be achieved by "raise Exception('...') from None"
            attr = None

        if attr is None:
            raise NotImplementedError(
                'The {} function has not been implemented'.format(ufunc)
            )

        if kwargs:
            warnings.warn('**kwargs, {}, are currently not supported'
                          .format(kwargs), stacklevel=2)

        case = len(inputs)
        if case == 1:
            pass  # Must be an UncertainArray
        elif case == 2:
            # At least 1 of the inputs must be an UncertainArray
            # If an input is not an ndarray then convert it to be an ndarray
            not0 = not isinstance(inputs[0], np.ndarray)
            if not0 or not isinstance(inputs[1], np.ndarray):
                # A tuple cannot be modified
                # This does not create a copy of the items
                inputs = list(inputs)
                # convert the input that is not an ndarray
                convert, keep = (0, 1) if not0 else (1, 0)
                if isinstance(inputs[convert], (Number, UncertainReal, UncertainComplex)):
                    inputs[convert] = np.full(inputs[keep].shape, inputs[convert], dtype=object)
                else:
                    inputs[convert] = np.asarray(inputs[convert], dtype=object)

            self._broadcasted_shape = None
            if inputs[0].shape != inputs[1].shape:
                broadcasted = np.broadcast(*inputs)
                inputs = broadcasted.iters
                self._broadcasted_shape = broadcasted.shape

        else:
            assert False, 'Should not occur: __array_ufunc__ received {} inputs'.format(case)

        return attr(*inputs)

    def __repr__(self):
        # Use the numpy formatting but hide the default dtype
        np_array_repr = np.array_repr(self)

        if self.dtype == object:
            # Truncate string from trailing ','
            i = np_array_repr.rfind(',')
            return np_array_repr[:i] + ')'
        else:
            return np_array_repr

    def __matmul__(self, other):
        # Implements the protocol used by the '@' operator defined in PEP 465.
        return matmul(self, other)

    def __rmatmul__(self, other):
        # Implements the protocol used by the '@' operator defined in PEP 465.
        return matmul(other, self)

    def _matmul(self, *inputs):
        # np.matmul became a ufunc in version 1.16.0
        return matmul(*inputs)

    def _create_empty(self, inputs=None, dtype=None, order='C'):
        if dtype is None:
            dtype = object
        shape = self.shape if self._broadcasted_shape is None else self._broadcasted_shape
        a = np.empty(shape, dtype=dtype, order=order)
        if inputs is None:
            return a, a.itemset, self.flat
        if len(inputs) == 1:
            return a, a.itemset, inputs[0].flat
        if isinstance(inputs[0], np.ndarray):
            return a, a.itemset, izip(inputs[0].flat, inputs[1].flat)
        # then the inputs are already broadcasted iterators
        return a, a.itemset, izip(*inputs)

    @property
    def label(self):
        """The label that was assigned to the array when it was created.

        **Example**::

            >>> current = la.uarray([ureal(0.57, 0.18), ureal(0.45, 0.12), ureal(0.68, 0.19)], label='amps')
            >>> current.label
            'amps'

        :rtype: :class:`str`
        """
        return self._label

    @property
    def real(self):
        """The result of applying the attribute ``real`` to each
        element in the array.

        **Example**::

            >>> a = la.uarray([ucomplex(1.2-0.5j, 0.6), ucomplex(3.2+1.2j, (1.4, 0.2)), ucomplex(1.5j, 0.9)])
            >>> a.real
            uarray([ureal(1.2,0.6,inf), ureal(3.2,1.4,inf),
                    ureal(0.0,0.9,inf)])

        :rtype: :class:`UncertainArray`
        """
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, item.real)
        return UncertainArray(arr)

    @property
    def imag(self):
        """The result of applying the attribute ``imag`` to each
        element in the array.

        **Example**::

            >>> a = la.uarray([ucomplex(1.2-0.5j, 0.6), ucomplex(3.2+1.2j, (1.4, 0.2)), ucomplex(1.5j, 0.9)])
            >>> a.imag
            uarray([ureal(-0.5,0.6,inf), ureal(1.2,0.2,inf),
                    ureal(1.5,0.9,inf)])

        :rtype: :class:`UncertainArray`
        """
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, item.imag)
        return UncertainArray(arr)

    @property
    def r(self):
        """The result of applying the attribute ``r`` to  each element in the array.

        **Example**::

            >>> a = la.uarray([ucomplex(1.2-0.5j, (1.2, 0.7, 0.7, 2.2)),
            ...                ucomplex(-0.2+1.2j, (0.9, 0.4, 0.4, 1.5))])
            >>> a.r
            uarray([0.43082021842766455, 0.34426518632954817])

        :rtype: :class:`UncertainArray`
        """
        arr, itemset, iterator = self._create_empty(dtype=None)
        for i, item in enumerate(iterator):
            itemset(i, item.r)
        return UncertainArray(arr)

    @property
    def x(self):
        """The result of :func:`~.core.value` for each element in the array.

        **Example**::

            >>> a = la.uarray([0.57, ureal(0.45, 0.12), ucomplex(1.1+0.68j, 0.19)])
            >>> a.x
            uarray([0.57, 0.45, (1.1+0.68j)])

        :rtype: :class:`UncertainArray`
        """
        return self.value()

    def value(self):
        """The result of :func:`~.core.value` for each element in the array.

        **Example**::

            >>> a = la.uarray([0.57, ureal(0.45, 0.12), ucomplex(1.1+0.68j, 0.19)])
            >>> a.value()
            uarray([0.57, 0.45, (1.1+0.68j)])

        :rtype: :class:`UncertainArray`
        """
        # Note: in the future we might allow different `dtype` values.
        # However, this needs some thought. Should `dtype=float`
        # return complex numbers as a pair of reals, for example?
        # What are the most likely use-cases?
        # :param dtype: The data type of the returned array.
        # :type dtype: :class:`numpy.dtype`
        arr, itemset, iterator = self._create_empty(dtype=None)
        for i, item in enumerate(iterator):
            itemset(i, value(item))
        return UncertainArray(arr)

    @property
    def u(self):
        """The result of :func:`~.core.uncertainty` for each element in the array.

        **Example**::

            >>> r = la.uarray([ureal(0.57, 0.18), ureal(0.45, 0.12), ureal(0.68, 0.19)])
            >>> r.u
            uarray([0.18, 0.12, 0.19])
            >>> c = la.uarray([ucomplex(1.2-0.5j, 0.6), ucomplex(3.2+1.2j, (1.4, 0.2)), ucomplex(1.5j, 0.9)])
            >>> c.u
            uarray([StandardUncertainty(real=0.6, imag=0.6),
                   StandardUncertainty(real=1.4, imag=0.2),
                   StandardUncertainty(real=0.9, imag=0.9)])

        :rtype: :class:`UncertainArray`
        """
        return self.uncertainty()

    def uncertainty(self):
        """The result of :func:`~.core.uncertainty` for each element in the array.

        **Example**::

            >>> r = la.uarray([ureal(0.57, 0.18), ureal(0.45, 0.12), ureal(0.68, 0.19)])
            >>> r.uncertainty()
            uarray([0.18, 0.12, 0.19])
            >>> c = la.uarray([ucomplex(1.2-0.5j, 0.6), ucomplex(3.2+1.2j, (1.4, 0.2)), ucomplex(1.5j, 0.9)])
            >>> c.uncertainty()
            uarray([StandardUncertainty(real=0.6, imag=0.6),
                   StandardUncertainty(real=1.4, imag=0.2),
                   StandardUncertainty(real=0.9, imag=0.9)])

        :rtype: :class:`UncertainArray`
        """
        # Note: in the future we might allow different `dtype` values.
        # However, we need to consider the use-cases carefully.
        # :param dtype: The data type of the returned array.
        # :type dtype: :class:`numpy.dtype`
        arr, itemset, iterator = self._create_empty(dtype=None)
        for i, item in enumerate(iterator):
            itemset(i, uncertainty(item))
        return UncertainArray(arr)

    @property
    def v(self):
        """The result of :func:`~.core.variance` for each element in the array.

        **Example**::

            >>> r = la.uarray([ureal(0.57, 0.18), ureal(0.45, 0.12), ureal(0.68, 0.19)])
            >>> r.v
            uarray([0.0324, 0.0144, 0.0361])
            >>> c = la.uarray([ucomplex(1.2-0.5j, 0.6), ucomplex(3.2+1.2j, (1.5, 0.5)), ucomplex(1.5j, 0.9)])
            >>> c.v
            uarray([VarianceCovariance(rr=0.36, ri=0.0, ir=0.0, ii=0.36),
                    VarianceCovariance(rr=2.25, ri=0.0, ir=0.0, ii=0.25),
                    VarianceCovariance(rr=0.81, ri=0.0, ir=0.0, ii=0.81)])

        :rtype: :class:`UncertainArray`
        """
        return self.variance()

    def variance(self):
        """The result of :func:`~.core.variance` for each element in the array.

        **Example**::

            >>> r = la.uarray([ureal(0.57, 0.18), ureal(0.45, 0.12), ureal(0.68, 0.19)])
            >>> r.variance()
            uarray([0.0324, 0.0144, 0.0361])
            >>> c = la.uarray([ucomplex(1.2-0.5j, 0.6), ucomplex(3.2+1.2j, (1.5, 0.5)), ucomplex(1.5j, 0.9)])
            >>> c.variance()
            uarray([VarianceCovariance(rr=0.36, ri=0.0, ir=0.0, ii=0.36),
                    VarianceCovariance(rr=2.25, ri=0.0, ir=0.0, ii=0.25),
                    VarianceCovariance(rr=0.81, ri=0.0, ir=0.0, ii=0.81)])

        :rtype: :class:`UncertainArray`
        """
        # Note: in the future we might allow different `dtype` values.
        # However, we need to consider the use-cases carefully.
        # :param dtype: The data type of the returned array.
        # :type dtype: :class:`numpy.dtype`
        arr, itemset, iterator = self._create_empty(dtype=None)
        for i, item in enumerate(iterator):
            itemset(i, variance(item))
        return UncertainArray(arr)

    @property
    def df(self):
        """The result of :func:`~.core.dof` for each element in the array.

        **Example**::

            >>> a = la.uarray([ureal(6, 2, df=3), ureal(4, 1, df=4), ureal(5, 3, df=7), ureal(1, 1)])
            >>> a.df
            uarray([3.0, 4.0, 7.0, inf])

        :rtype: :class:`UncertainArray`
        """
        return self.dof()

    def dof(self):
        """The result of :func:`~.core.dof` for each element in the array.

        **Example**::

            >>> a = la.uarray([ureal(6, 2, df=3), ureal(4, 1, df=4), ureal(5, 3, df=7), ureal(1, 1)])
            >>> a.dof()
            uarray([3.0, 4.0, 7.0, inf])

        :rtype: :class:`UncertainArray`
        """
        arr, itemset, iterator = self._create_empty(dtype=None)
        for i, item in enumerate(iterator):
            itemset(i, dof(item))
        return UncertainArray(arr)

    def sensitivity(self, x):
        """The result of :func:`~.reporting.sensitivity` for each element in the array.

        :rtype: :class:`UncertainArray`
        """
        # Note, there is a case for introducing `dtype` or some other parameter.
        # The return types for complex cases may be multivariate.

        # `_create_empty()` handles only ndarray-like sequences
        if not isinstance(x, np.ndarray):
            x = np.asarray(x)

        arr, itemset, iterator = self._create_empty((self, x))
        for i, (y, x) in enumerate(iterator):
            itemset(i, y.sensitivity(x))
        return UncertainArray(arr)

    def u_component(self, x):
        """The result of :func:`~.reporting.u_component` for each element in the array.

        :rtype: :class:`UncertainArray`
        """
        # Note, there is a case for introducing `dtype` or some other parameter.
        # The return types for complex cases may be multivariate.

        # `_create_empty()` handles only ndarray-like sequences
        if not isinstance(x, np.ndarray):
            x = np.asarray(x)

        arr, itemset, iterator = self._create_empty((self, x))
        for i, (y, x) in enumerate(iterator):
            itemset(i, y.u_component(x))
        return UncertainArray(arr)

    def conjugate(self):
        """The result of applying the attribute ``conjugate`` to each element in the array.

        **Example**::

            >>> a = la.uarray([ucomplex(1.2-0.5j, 0.6), ucomplex(3.2+1.2j, (1.4, 0.2)), ucomplex(1.5j, 0.9)])
            >>> a.conjugate()
            uarray([ucomplex((1.2+0.5j), u=[0.6,0.6], r=0.0, df=inf),
                    ucomplex((3.2-1.2j), u=[1.4,0.2], r=0.0, df=inf),
                    ucomplex((0-1.5j), u=[0.9,0.9], r=0.0, df=inf)])

        :rtype: :class:`UncertainArray`
        """
        # override this method because I wanted to create a custom __doc__
        return self._conjugate()

    def _conjugate(self, *ignore):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, item.conjugate())
        return UncertainArray(arr)

    def _positive(self, *ignore):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, +item)
        return UncertainArray(arr)

    def _negative(self, *ignore):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, -item)
        return UncertainArray(arr)

    def _add(self, *inputs):
        arr, itemset, iterator = self._create_empty(inputs)
        for i, (a, b) in enumerate(iterator):
            itemset(i, a + b)
        return UncertainArray(arr)

    def _subtract(self, *inputs):
        arr, itemset, iterator = self._create_empty(inputs)
        for i, (a, b) in enumerate(iterator):
            itemset(i, a - b)
        return UncertainArray(arr)

    def _multiply(self, *inputs):
        arr, itemset, iterator = self._create_empty(inputs)
        for i, (a, b) in enumerate(iterator):
            itemset(i, a * b)
        return UncertainArray(arr)

    def _divide(self, *inputs):
        return self._true_divide(*inputs)

    def _true_divide(self, *inputs):
        arr, itemset, iterator = self._create_empty(inputs)
        for i, (a, b) in enumerate(iterator):
            itemset(i, a / b)
        return UncertainArray(arr)

    def _power(self, *inputs):
        arr, itemset, iterator = self._create_empty(inputs)
        for i, (a, b) in enumerate(iterator):
            itemset(i, a ** b)
        return UncertainArray(arr)

    def _exp(self, *ignore):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, exp(item))
        return UncertainArray(arr)

    def _log(self, *ignore):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, log(item))
        return UncertainArray(arr)

    def _log10(self, *ignore):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, log10(item))
        return UncertainArray(arr)

    def _sqrt(self, *ignore):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, sqrt(item))
        return UncertainArray(arr)

    def _cos(self, *ignore):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, cos(item))
        return UncertainArray(arr)

    def _sin(self, *ignore):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, sin(item))
        return UncertainArray(arr)

    def _tan(self, *ignore):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, tan(item))
        return UncertainArray(arr)

    def _arccos(self, *ignore):
        return self._acos()

    def _acos(self):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, acos(item))
        return UncertainArray(arr)

    def _arcsin(self, *ignore):
        return self._asin()

    def _asin(self):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, asin(item))
        return UncertainArray(arr)

    def _arctan(self, *ignore):
        return self._atan()

    def _atan(self):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, atan(item))
        return UncertainArray(arr)

    def _arctan2(self, *inputs):
        return self._atan2(inputs[1])

    def _atan2(self, *inputs):
        arr, itemset, iterator = self._create_empty((self, inputs[0]))
        for i, (a, b) in enumerate(iterator):
            itemset(i, atan2(a, b))
        return UncertainArray(arr)

    def _sinh(self, *ignore):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, sinh(item))
        return UncertainArray(arr)

    def _cosh(self, *ignore):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, cosh(item))
        return UncertainArray(arr)

    def _tanh(self, *ignore):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, tanh(item))
        return UncertainArray(arr)

    def _arccosh(self, *ignore):
        return self._acosh()

    def _acosh(self):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, acosh(item))
        return UncertainArray(arr)

    def _arcsinh(self, *ignore):
        return self._asinh()

    def _asinh(self):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, asinh(item))
        return UncertainArray(arr)

    def _arctanh(self, *ignore):
        return self._atanh()

    def _atanh(self):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, atanh(item))
        return UncertainArray(arr)

    def _square(self, *ignore):
        return self._mag_squared()

    def _mag_squared(self):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, mag_squared(item))
        return UncertainArray(arr)

    def _magnitude(self):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, magnitude(item))
        return UncertainArray(arr)

    def _phase(self):
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            itemset(i, phase(item))
        return UncertainArray(arr)

    def _intermediate(self, labels):
        # Default second argument of calling function is `None`
        if labels is None:
            arr, itemset, iterator = self._create_empty()
            for i, x in enumerate(iterator):
                itemset(i, result(x))
        else:
            # `_create_empty()` handles only ndarray-like sequences
            if not is_sequence(labels):
                # Add index notation to the label base
                labels = [
                    "{}[{}]".format(labels, i)
                    for i in xrange(self.size)
                ]

            labels = np.asarray(labels)
            arr, itemset, iterator = self._create_empty((self, labels))
            for i, (x, lbl) in enumerate(iterator):
                itemset(i, result(x, lbl))

        return UncertainArray(arr)

    def _equal(self, *inputs):
        arr, itemset, iterator = self._create_empty(inputs, dtype=bool)
        for i, (a, b) in enumerate(iterator):
            itemset(i, a == b)
        return arr

    def _not_equal(self, *inputs):
        arr, itemset, iterator = self._create_empty(inputs, dtype=bool)
        for i, (a, b) in enumerate(iterator):
            itemset(i, a != b)
        return arr

    def _less(self, *inputs):
        arr, itemset, iterator = self._create_empty(inputs, dtype=bool)
        for i, (a, b) in enumerate(iterator):
            itemset(i, a < b)
        return arr

    def _less_equal(self, *inputs):
        arr, itemset, iterator = self._create_empty(inputs, dtype=bool)
        for i, (a, b) in enumerate(iterator):
            itemset(i, a <= b)
        return arr

    def _greater(self, *inputs):
        arr, itemset, iterator = self._create_empty(inputs, dtype=bool)
        for i, (a, b) in enumerate(iterator):
            itemset(i, a > b)
        return arr

    def _greater_equal(self, *inputs):
        arr, itemset, iterator = self._create_empty(inputs, dtype=bool)
        for i, (a, b) in enumerate(iterator):
            itemset(i, a >= b)
        return arr

    def _maximum(self, *inputs):
        arr, itemset, iterator = self._create_empty(inputs)
        for i, (a, b) in enumerate(iterator):
            if _isnan(a):
                itemset(i, a)
            elif _isnan(b):
                itemset(i, b)
            elif a > b:
                itemset(i, a)
            else:
                itemset(i, b)
        return UncertainArray(arr)

    def _minimum(self, *inputs):
        arr, itemset, iterator = self._create_empty(inputs)
        for i, (a, b) in enumerate(iterator):
            if _isnan(a):
                itemset(i, a)
            elif _isnan(b):
                itemset(i, b)
            elif a < b:
                itemset(i, a)
            else:
                itemset(i, b)
        return UncertainArray(arr)

    def _logical_and(self, *inputs):
        arr, itemset, iterator = self._create_empty(inputs, dtype=object)
        for i, (a, b) in enumerate(iterator):
            itemset(i, a and b)
        return UncertainArray(arr)

    def _logical_or(self, *inputs):
        arr, itemset, iterator = self._create_empty(inputs, dtype=object)
        for i, (a, b) in enumerate(iterator):
            itemset(i, a or b)
        return UncertainArray(arr)

    def _logical_xor(self, *inputs):
        raise TypeError(
            "Boolean bitwise operations are not defined for `UncertainArray`"
        )
        # arr, itemset, iterator = self._create_empty(inputs, dtype=bool)
        # for i, (a, b) in enumerate(iterator):
        #     itemset(i, bool(a) ^ bool(b))
        # return arr

    def _logical_not(self, *inputs):
        arr, itemset, iterator = self._create_empty(inputs, dtype=bool)
        for i, item in enumerate(iterator):
            itemset(i, not bool(item))
        return arr

    def _isinf(self, *inputs):
        arr, itemset, iterator = self._create_empty(inputs, dtype=bool)
        for i, item in enumerate(iterator):
            itemset(i, _isinf(item))
        return arr

    def _isnan(self, *inputs):
        arr, itemset, iterator = self._create_empty(inputs, dtype=bool)
        for i, item in enumerate(iterator):
            itemset(i, _isnan(item))
        return arr

    def _isfinite(self, *inputs):
        arr, itemset, iterator = self._create_empty(inputs, dtype=bool)
        for i, item in enumerate(iterator):
            itemset(i, not (_isnan(item) or _isinf(item)))
        return arr

    def _reciprocal(self, *inputs):
        arr, itemset, iterator = self._create_empty(inputs)
        for i, item in enumerate(iterator):
            itemset(i, 1.0/item)
        return UncertainArray(arr)

    def _absolute(self, *inputs):
        arr, itemset, iterator = self._create_empty(inputs)
        for i, item in enumerate(iterator):
            itemset(i, abs(item))
        return UncertainArray(arr)

    def copy(self, order='C'):
        arr, itemset, iterator = self._create_empty(order=order)
        for i, item in enumerate(iterator):
            itemset(i, +item)
        return UncertainArray(arr, label=self.label)

    def round(self, decimals=0, **kwargs):
        digits = kwargs.get('digits', decimals)
        df_decimals = kwargs.get('df_decimals', digits)
        arr, itemset, iterator = self._create_empty()
        for i, item in enumerate(iterator):
            try:
                itemset(i, item._round(digits, df_decimals))
            except AttributeError:
                try:
                    itemset(i, round(item, digits))
                except TypeError:
                    itemset(i, complex(round(item.real, digits), round(item.imag, digits)))
        return UncertainArray(arr)

    def sum(self, *args, **kwargs):
        raise TypeError(
            "`sum` is not defined for `UncertainArray`"
        )
        # return UncertainArray(np.asarray(self).sum(*args, **kwargs))

    def mean(self, *args, **kwargs):
        raise TypeError(
            "`mean` is not defined for `UncertainArray`"
        )
        # return UncertainArray(np.asarray(self).mean(*args, **kwargs))

    def std(self, *args, **kwargs):
        # If this is to be implemented we need to be clear about
        # what is calculated. This will not be an uncertain-number
        # calculation, it will take the values of a sample of uncertain
        # numbers and evaluate the SD. This will probably be clearer
        # if the function is in the `type_a` module.
        # Note we would also want a similar function to calculate
        # the standard error (ie the type-A uncertainty).
        raise TypeError(
            "`std` is not defined for `UncertainArray`"
        )
        # return UncertainArray(np.asarray(self).std(*args, **kwargs))

    def var(self, *args, **kwargs):
        # If this is to be implemented we need to be clear about
        # what is calculated. This will not be an uncertain-number
        # calculation, it will take the values of a sample of uncertain
        # numbers and evaluate the SD. This will probably be clearer
        # if the function is in the `type_a` module.
        # Note we would also want a similar function to calculate
        # the standard variance (ie the type-A uncertainty squared).
        raise TypeError(
            "`var` is not defined for `UncertainArray`"
        )
        # return UncertainArray(np.asarray(self).var(*args, **kwargs))

    def max(self, *args, **kwargs):
        raise TypeError(
            "`max` is not defined for `UncertainArray`"
        )
        # return UncertainArray(np.asarray(self).max(*args, **kwargs))

    def min(self, *args, **kwargs):
        raise TypeError(
            "`min` is not defined for `UncertainArray`"
        )
        # return UncertainArray(np.asarray(self).min(*args, **kwargs))

    def trace(self, *args, **kwargs):
        raise TypeError(
            "`trace` is not defined for `UncertainArray`"
        )
        # return UncertainArray(np.asarray(self).trace(*args, **kwargs))

    def cumprod(self, *args, **kwargs):
        # numpy catches ``TypeError`` and uses its
        # internal implementation of this method
        raise RuntimeError(
            "`cumprod` is not defined for `UncertainArray`"
        )
        # return UncertainArray(np.asarray(self).cumprod(*args, **kwargs))

    def cumsum(self, *args, **kwargs):
        # numpy catches ``TypeError`` and uses its
        # internal implementation of this method
        raise RuntimeError(
            "`cumsum` is not defined for `UncertainArray`"
        )
        # return UncertainArray(np.asarray(self).cumsum(*args, **kwargs))

    def prod(self, *args, **kwargs):
        raise TypeError(
            "`prod` is not defined for `UncertainArray`"
        )
        # return UncertainArray(np.asarray(self).prod(*args, **kwargs))

    def ptp(self, *args, **kwargs):
        raise TypeError(
            "`ptp` is not defined for `UncertainArray`"
        )
        # return UncertainArray(np.asarray(self).ptp(*args, **kwargs))

    def any(self, *args, **kwargs):
        raise TypeError(
            "`any` is not defined for `UncertainArray`"
        )
        # return UncertainArray(np.asarray(self, dtype=bool).any(*args, **kwargs))

    def all(self, *args, **kwargs):
        raise TypeError(
            "`all` is not defined for `UncertainArray`"
        )
        # return UncertainArray(np.asarray(self, dtype=bool).all(*args, **kwargs))


# Allows pickle to understand the class name 'uarray'
uarray = UncertainArray



import unittest

from GTC import ucomplex
from GTC.lib import UncertainReal

TOL = 1E-13 

#----------------------------------------------------------------------------
class TestContext(unittest.TestCase):

    def test_construction(self):        
        # import here to get the tests to work with Python 2.7 on linux
        from GTC.context import _context

        x_value = 1.2
        x_u = 0.5
        x_dof = 6
        x1 = UncertainReal._elementary(x_value,x_u,x_dof,None,independent=True)
        x2 = UncertainReal._elementary(x_value,x_u,x_dof,None,independent=True)

        # uid's must be in order
        self.assertTrue( x1._node.uid < x2._node.uid )
        self.assertEqual(x_dof, x1._node.df )
        self.assertEqual(x_dof, x2._node.df )

        self.assertTrue( _context._registered_leaf_nodes[x1._node.uid].df == x_dof )
        self.assertTrue( _context._registered_leaf_nodes[x2._node.uid].df == x_dof )

        # illegal dof is checked when the object is created
        self.assertRaises(
            ValueError,
            UncertainReal._elementary,x_value,x_u,0,None,False
        )

    def test_invalid_ucomplex_creation(self):
        # RuntimeError: covariance elements not equal: None and 1
        self.assertRaises(TypeError,ucomplex,1 + 0j, (1, None, 1, 1))
        # RuntimeError: covariance elements not equal: 999999j and 1
        self.assertRaises(TypeError,ucomplex,1 + 0j, (1, 999999j, 1, 1))

        
#============================================================================
if(__name__== '__main__'):

    unittest.main()    # Runs all test methods starting with 'test'


"""
Class
-----

    An :class:`Archive` is used to store and retrieve uncertain numbers. 
    
Module contents
---------------

"""
import itertools
import sys
PY2 = bool( sys.version_info[0] == 2 )

from GTC.lib import (
    UncertainComplex,
    UncertainReal
)
from GTC.vector import Vector 
from GTC import context

__all__ = (
    'Archive',
)

#============================================================================
# When an archive is prepared for storage, uncertain number objects 
# are translated into the following simple representations that
# Python will pickle.
#
class LeafNode(object):
    def __init__(self,node):
        
        # If called during freezing, a
        # Leaf object will be provided.
        # If called after a JSON record has 
        # been restored, a dict will be provided.
        
        if isinstance( node, dict ): 
            # JSON record
            
            # A uid must be hashable
            self.uid = tuple(node['uid'])
            
            # 'None' must be restored to an object
            label = node['label']
            if label != "None" : 
                self.label = label 
            else:
                self.label = None
                
            self.u = float( node['u'] )
            self.df = float( node['df'] )
            
            self.independent = bool( node['independent'] ) 
            
            if 'complex' in node:
                self.complex = ( 
                    eval(node['complex'][0]), 
                    eval(node['complex'][1]) 
                )
            if 'correlation' in node:
                items = node['correlation'].iteritems() if PY2 else node['correlation'].items()
                self.correlation = {
                    eval(x_i) : r_i 
                        for x_i,r_i in items
                }
            if 'ensemble' in node:
                self.ensemble = frozenset( 
                    eval(i) for i in node['ensemble'] 
                )                               
        else:    
            # Leaf object        
            self.uid = node.uid
            self.label = node.label 
            self.u = node.u 
            self.df = node.df 
            self.independent = node.independent 
            if hasattr(node,'complex'):
                self.complex = node.complex
            if hasattr(node,'correlation'):
                self.correlation = node.correlation.items()
                if not PY2:
                    # In Python 2 items() returns a list but
                    # in Python 3 a dict_items object is returned.
                    # Since it is not possible to pickle a
                    # dict_items object we convert dict_items
                    # to be a list.
                    self.correlation = list(self.correlation)
            if hasattr(node,'ensemble'):
                self.ensemble = frozenset( node.ensemble )
        
class ElementaryReal(object):
    def __init__(self,x,uid):
        self.x = x
        self.uid = uid

class IntermediateReal(object):
    def __init__(self,value,u_components,d_components,i_components,label,uid):
        self.value = value
        self.u_components = u_components    
        self.d_components = d_components     
        self.i_components = i_components    
        self.label = label
        self.uid = uid
    
class Complex(object):
    def __init__(self,n_re,n_im,label):
        self.n_re = n_re
        self.n_im = n_im
        self.label = label
            
#----------------------------------------------------------------------------
# """
# Archive plays several roles related to storage of uncertain numbers.

# When uncertain numbers are stored, an archive is used to marshal
# information related to the selected uncertain numbers. This 
# ensures that when they are retrieved from storage, later in another Python  
# process, their behaviour will be the same as it was in the original process.

# This information is collected after specific uncertain numbers have been 
# identified. It is referred to as 'freezing' the archive. Freezing also 
# involves a transformation of the archive object contents into a form that 
# is suitable for generic storage in different formats. In this way, the 
# archive represents an intermediate staging point, which can be 
# followed by different format-specific storage operations.

# When a stored archive is to be retrieved, the reverse process occurs.
# First, format-specific operations must be used to recreate an archive 
# in its frozen state. Then the archive will be 'thawed', which will 
# create various GTC objects and restore the necessary environment for 
# the specific uncertain numbers that were saved initially. Finally, 
# these uncertain numbers may be restored (on demand).
# """
class Archive(object):
    """
    An :class:`Archive` helps to store and retrieve uncertain numbers,
    so that they can be used in later calculations. 
    
    A particular :class:`Archive` object can either be used to prepare 
    a record of uncertain numbers for storage, or to retrieve a stored record. 
        
    """
    def __init__(self):

        # `self._tagged_reals` contains information about every
        # real uncertain number destined for storage: a pair 
        # of entries is included when a complex UN is tagged.
        #
        self._tagged = {}           # name->object-ref pairs
        self._tagged_reals = {}     # name->object-ref pairs
        
        # Filled by add() and then used when freezing to
        # associate the uid's of intermediate components with UNs.
        self._uid_to_intermediate = {}
        
        self._extract = False   # initially in write-only mode

    def keys(self):
        """Return a list of names 
        """
        return self._tagged.keys()

    def iterkeys(self):
        """Return an iterator for names 
        """
        if PY2:
            return self._tagged.iterkeys()
        else:
            return self._tagged.keys()

    def values(self):
        """Return a list of uncertain numbers 
        """
        return self._tagged.values() 
        
    def itervalues(self):
        """Return an iterator for uncertain numbers 
        """
        if PY2:
            return self._tagged.itervalues()
        else:
            return self._tagged.values()

    def items(self):
        """Return a list of name -to- uncertain-number pairs 
        """
        return self._tagged.items()
        
    def iteritems(self):
        """Return an iterator of name -to- uncertain-number pairs 
        """
        if PY2:
            return self._tagged.iteritems()
        else:
            return self._tagged.items()

    def __len__(self):
        """Return the number of entries 
        """
        return len(self._tagged)

    def _setitem(self,key,value):
        """Add a name -to- uncertain-number pair
        
        """
        if key in self._tagged:
            raise RuntimeError(
                "'{!s}' is already in use".format(key)
            )
        else:
            if isinstance(value,UncertainReal):
                if key in self._tagged_reals:
                    raise RuntimeError(
                        "'{!s}' is being used as a name-tag".format(key)
                    )
                    
                if not value.is_elementary:
                    try:
                        uid = value.real._node.uid
                    except AttributeError:
                        raise RuntimeError(
                            "uncertain number labelled '{}' is not declared intermediate".format(key)
                        )
                        
                    self._uid_to_intermediate[uid] = value.real            
                
                self._tagged_reals[key] = value

            elif isinstance(value,UncertainComplex):
            
                n_re = "{!s}_re".format(key)
                if n_re in self._tagged_reals:
                    raise RuntimeError(
                        "'{!s}' is being used as a name-tag".format(n_re)
                    )

                n_im = "{!s}_im".format(key)
                if n_im in self._tagged_reals:
                    raise RuntimeError(
                        "'{!s}' is being used as a name-tag".format(n_im)
                    )
                    
                self._tagged_reals[n_re] = value.real
                self._tagged_reals[n_im] = value.imag
                
                if not value.is_elementary:
                    try:
                        uid_r = value.real._node.uid
                        uid_i = value.imag._node.uid
                    except AttributeError:
                        raise RuntimeError(
                            "uncertain number labelled '{}' is not declared intermediate".format(key)
                        )
                        
                    self._uid_to_intermediate[uid_r] = value.real
                    self._uid_to_intermediate[uid_i] = value.imag
                        
            else:
                raise RuntimeError(
                    "'{!r}' cannot be archived: wrong type"
                )
                
            self._tagged[key] = value

    def __setitem__(self,key,value):
        """
        Add an uncertain number to the archive
        
        **Example**::
        
            >>> a = Archive()
            >>> x = ureal(1,1)
            >>> y = ureal(2,1)
            >>> a['x'] = x
            >>> a['fred'] = y
            
        """
        if self._extract:
            raise RuntimeError('This archive is read-only!')
        else:
            self._setitem(key,value)


    def add(self,**kwargs):
        """Add entries to an archive.
        
        Each entry is given a name that identifies it within the archive.

        **Example**
        
        .. code-block:: pycon   
        
            >>> a = pr.Archive()
            >>> x = ureal(1,1)
            >>> y = ureal(2,1)
            >>> z = ureal(20,1)
            >>> a.add(x=x,fred=y)

            # Entries can also be added using the name as a key       
            >>> a['z'] = z

        .. invisible-code-block: pycon
        
            >>> import tempfile
            >>> f = open(tempfile.gettempdir() + '/GTC-archive-test.gar', 'wb')

        Here ``f`` is a file stream opened in mode 'wb':
        
        .. code-block:: pycon   
        
            >>> pr.dump(f, a)
            >>> f.close()
   

        """
        if self._extract:
            raise RuntimeError('This archive is write-only!')

        items = kwargs.iteritems() if PY2 else kwargs.items()
        for key,value in items:
            self._setitem(key,value)

    def _getitem(self,key):
        """
        """
        try:
            value = self._tagged[key]
        except KeyError:
            raise RuntimeError(
                "'{!s}' not found".format(key)
            )

        return value

    def __getitem__(self,key):
        """Extract an uncertain number

        `key` - the name of the archived number
        
        """
        if not self._extract:
            raise RuntimeError('This archive is write-only!')
        else:
            return self._getitem(key)

    def extract(self,*args):
        """
        Extract uncertain numbers by name

        :arg args: names of uncertain numbers stored in the archive
        
        If just one name is provided, a single uncertain number is returned.
        Otherwise a sequence of uncertain numbers is returned.
        
        **Example**
        
        Continuing the example in :meth:`~.Archive.add`, but in a different 
        Python session, ``f`` is now a file stream opened in 'rb' mode:

        .. invisible-code-block: pycon

            >>> import tempfile
            >>> f = open(tempfile.gettempdir() + '/GTC-archive-test.gar', 'rb')
            
        .. code-block:: pycon
        
            >>> a = pr.load(f)
            >>> f.close()
            
            >>> a.extract('fred')
            ureal(2.0,1.0,inf)
            >>> x, fred = a.extract('x','fred')
            >>> x
            ureal(1.0,1.0,inf)            
 
            # Entries can also be extracted using the name as a key
            >>> a['z']
            ureal(20.0,1.0,inf)

        .. invisible-code-block: pycon
            
            >>> import os, tempfile
            >>> os.remove(tempfile.gettempdir() + '/GTC-archive-test.gar')
  
        """        
        if not self._extract:
            raise RuntimeError('This archive is read-only!')
        
        lst = [ self._getitem(n) for n in args ]
            
        return lst if len(lst) > 1 else lst[0]

    # -----------------------------------------------------------------------
    def _freeze(self):
        """Prepare archive for storage
        
        NB after freezing, the archive object is immutable.
        
        """    
        if not len(self):
            raise RuntimeError(
                "There is nothing in the archive!"
            )

        values = self._tagged_reals.itervalues() if PY2 else self._tagged_reals.values()
        
        self._leaf_nodes = {
            n_i.uid  : LeafNode(n_i)
                for un in values
                    for n_i in itertools.chain(
                        un._u_components.iterkeys(),
                        un._d_components.iterkeys()
                    )
        }                      
                        
        # -------------------------------------
        # Intermediate real uncertain numbers
        #
        # All elementary influences of intermediate nodes 
        # have been found above and will be archived. 
        # However, intermediate influences may not have 
        # been tagged, in which case they are not archived.
        values = self._tagged_reals.itervalues() if PY2 else self._tagged_reals.values()
        _intermediate_node_to_uid = {
            v._node: v._node.uid 
            for v in values
                if not v.is_elementary
        }
                
        # Use this to recreate intermediate nodes in _thaw
        self._intermediate_uids = {
            n_i.uid : (n_i.label,n_i.u,n_i.df)
            for n_i in _intermediate_node_to_uid
        }
        
        # -------------------------------------------------------------------
        # Convert tagged objects into a standard form for storage 
        #
        items = self._tagged.iteritems() if PY2 else self._tagged.items()
        for n,obj in items:
            if obj.is_elementary:
                if isinstance(obj,UncertainReal):
                    tagged = ElementaryReal(
                        x=obj.x,
                        uid=obj._node.uid
                    )
                    self._tagged[n] = tagged
                    self._tagged_reals[n] = tagged
                    
                elif isinstance(obj,UncertainComplex):
                    re = ElementaryReal(
                        x=obj.real.x,
                        uid=obj.real._node.uid
                    )
                    im = ElementaryReal(
                        x=obj.imag.x,
                        uid=obj.imag._node.uid
                    )
                    n_re = "{}_re".format(n)
                    self._tagged_reals[n_re] = re
                    
                    n_im = "{}_im".format(n)
                    self._tagged_reals[n_im] = im
                    
                    self._tagged[n] = Complex(
                        n_re=n_re,
                        n_im=n_im,
                        label = obj.label
                    )
                else:
                    assert False, 'unexpected'
            else:
                if isinstance(obj,UncertainReal):
                    un = IntermediateReal(
                        value = obj.x,
                        u_components = _vector_index_to_uid( 
                            obj._u_components ),
                        d_components = _vector_index_to_uid( 
                            obj._d_components ),
                        i_components = _ivector_index_to_uid(
                            obj._i_components,
                            _intermediate_node_to_uid
                        ),
                        label = obj.label,
                        uid = obj._node.uid
                    )
                    self._tagged[n] = un
                    self._tagged_reals[n] = un
                    
                elif isinstance(obj,UncertainComplex):
                    re = IntermediateReal(
                        value = obj.real.x,
                        u_components = _vector_index_to_uid(
                            obj.real._u_components ),
                        d_components = _vector_index_to_uid( 
                            obj.real._d_components ),
                        i_components = _ivector_index_to_uid(
                            obj.real._i_components,
                            _intermediate_node_to_uid
                        ),
                        label = obj.real.label,
                        uid = obj.real._node.uid,
                    )
                        
                    n_re = "{}_re".format(n)
                    self._tagged_reals[n_re] = re
                    
                    im = IntermediateReal(
                        value = obj.imag._x,
                        u_components = _vector_index_to_uid( 
                            obj.imag._u_components ),
                        d_components = _vector_index_to_uid( 
                            obj.imag._d_components ),
                        i_components = _ivector_index_to_uid(
                            obj.imag._i_components,
                            _intermediate_node_to_uid
                        ),
                        label = obj.imag.label,
                        uid = obj.imag._node.uid,
                    )

                    n_im = "{}_im".format(n)
                    self._tagged_reals[n_im] = im
                    
                    self._tagged[n] = Complex(
                        n_re=n_re,
                        n_im=n_im,
                        label=obj.label
                    )
                else:
                    assert False,"unexpected"
                    
        # Python cannot pickle this
        del self._uid_to_intermediate 
        
    # -----------------------------------------------------------------------
    def _thaw(self):
        _leaf_nodes = dict()
        items = self._leaf_nodes.iteritems() if PY2 else self._leaf_nodes.items()
                        
        for uid_i,fl_i in items:
            l = context._context.new_leaf(
                uid_i, 
                fl_i.label, 
                fl_i.u, 
                fl_i.df, 
                fl_i.independent,
            )            
            if hasattr(fl_i,'complex'):
                l.complex = fl_i.complex 
            if hasattr(fl_i,'correlation'):
                l.correlation = dict( fl_i.correlation )
            if hasattr(fl_i,'ensemble'):
                l.ensemble = set( fl_i.ensemble )
            _leaf_nodes[uid_i] = l

        # Create the nodes associated with intermediate 
        # uncertain numbers. This must be done before the 
        # intermediate uncertain numbers are recreated.
        items = self._intermediate_uids.iteritems() if PY2 else self._intermediate_uids.items()
        
        # In v1.3.5, a df field was added to intermediate nodes. 
        # In previous versions it was absent. This shim function
        # will add `None` to the df attribute, which can then be
        # fixed once the uncertain number object is formed.
        # This is a temporary feature that will be removed after a few releases.
        # See also lib.py property df  
        shim_1_3_3 = lambda args: args + (None,) if len(args) == 2 else args
            
        _nodes = {
            uid: context._context.new_node(uid, *shim_1_3_3(args) )
                for uid, args in items
        }
            
        # When reconstructing, `_tagged` needs to be updated with 
        # the new uncertain numbers.
        #
        items = self._tagged.iteritems() if PY2 else self._tagged.items()
        for name,obj in items:
            if isinstance(obj,ElementaryReal):
                un = _builder(
                    name,
                    _leaf_nodes,
                    self._tagged_reals
                )                    
                self._tagged[name] = un

            elif isinstance(obj,IntermediateReal):
                un = _builder(
                    name,
                    _nodes,
                    self._tagged_reals
                )                    
                self._tagged[name] = un
                
            elif isinstance(obj,Complex):
                # This is an intermediate uncertain complex
                name_re = obj.n_re
                name_im = obj.n_im
                
                un_re = _builder(
                    name_re,
                    _nodes,
                    self._tagged_reals
                )
                un_im = _builder(
                    name_im,
                    _nodes,
                    self._tagged_reals
                )

                assert un_re.is_elementary == un_im.is_elementary
                
                unc = UncertainComplex(un_re,un_im)
                
                # An intermediate complex needs to 
                # link the nodes of its components 
                # (same as in `UncertainComplex._intermediate`)
                complex_id = (un_re._node.uid,un_im._node.uid)
                unc.real._node.complex = complex_id 
                unc.imag._node.complex = complex_id
        
                self._tagged[name] = unc
                
            else:
                assert False
                        
        # Change the archive status
        self._extract = True
        
#----------------------------------------------------------------------------
def _vector_index_to_uid(v):
    """
    Change the vector index from a node to a uid 
    
    """
    return Vector(
        index = [
            n_i.uid for n_i in v._index
        ],
        value = v._value
    )

#----------------------------------------------------------------------------
def _vector_index_to_node(v):
    """
    Change the vector index from a uid to a node
    
    """
    _nodes = context._context._registered_leaf_nodes
    return Vector(
        index = [
            _nodes[uid_i] for uid_i in v._index
        ],
        value = v._value
    )    

#----------------------------------------------------------------------------
def _ivector_index_to_uid(i_components,_intermediate_node_to_uid):
    """
    Return a Vector containing the uids of tagged intermediate UNs 
    
    """
    i_sequence = []
    v_sequence = []
    
    for i,v in i_components.iteritems():
        if i in _intermediate_node_to_uid:
            i_sequence.append( _intermediate_node_to_uid[i] )
            v_sequence.append(v)
            
    return Vector(index=i_sequence,value=v_sequence)
    
#----------------------------------------------------------------------------
def _ivector_index_to_node(i_components,_intermediate_node):
    """
    Return a Vector containing intermediate nodes as indices 
    
    """            
    return Vector(
        index= [
            _intermediate_node[uid_i]
            for uid_i in i_components.iterkeys()
            
        ],
        value=i_components._value
    )
    
#----------------------------------------------------------------------------
"""
Notes
-----

`tagged_real` is indexed by name and maps to an object to be constructed.

"""
def _builder(o_name,_nodes,_tagged_reals):
    """
    Construct an intermediate un object for `o_name`.
    
    """
    obj = _tagged_reals[o_name]
    
    if isinstance(obj,ElementaryReal):
        un = UncertainReal._archived_elementary(
            uid = obj.uid,
            x = obj.x
        )
        _tagged_reals[o_name] = un    
                
    elif isinstance(obj,IntermediateReal):                
            
        _node = _nodes[obj.uid] 
        
        un = UncertainReal(
            obj.value,
            _vector_index_to_node( obj.u_components ),
            _vector_index_to_node( obj.d_components ),
            _ivector_index_to_node( obj.i_components, _nodes ),
            _node,
            )
        
        _tagged_reals[o_name] = un

    else:
        assert False, "unexpected: {!r}".format(obj)

    return un

#============================================================================    
if __name__ == "__main__":
    import doctest
    from GTC import *  
    doctest.testmod( optionflags=doctest.NORMALIZE_WHITESPACE )



import unittest

from GTC import *
from GTC.lib import (
    real_ensemble,
    welch_satterthwaite,
    UncertainReal
)

from testing_tools import *

TOL = 1E-13 

#----------------------------------------------------------------------------
class SimpleWSCases(unittest.TestCase):
    def test(self):
        x = ureal(1,1,4)
        self.assertEqual(4,x.df)
        self.assertEqual(4,welch_satterthwaite(x)[1] )
        
        # product with zero values 
        x1 = ureal(0,1,4)
        x2 = ureal(0,1,3)
        y = x1 * x2
        self.assertEqual(0,y.u)
        self.assertTrue(nan is y.df)
        
        # Pathological case - not sure it can be created in practice
        x1 = ureal(1,1)
        x2 = UncertainReal._elementary(1,0,4,label=None,independent=True)
        x3 = UncertainReal._elementary(1,0,3,label=None,independent=True)
        y = x1 + x2 + x3
        self.assertEqual( len(y._u_components), 3 )
        self.assertTrue(inf is y.df)

#----------------------------------------------------------------------------
class SimpleComplexMagnitude(unittest.TestCase):
    def test(self):
        mag = magnitude( ucomplex(1+1j,[.5,.1,.1,.5],4) )
        self.assertEqual(4,mag.df)
        
#----------------------------------------------------------------------------
class GuideExampleH1(unittest.TestCase):

    def setUp(self):
        self.d_1 = ureal(0.0,5.8,24)
        self.d_2 = ureal(0.0,3.9,5)
        self.d_3 = ureal(0.0,6.7,8)

        self.alpha_s = ureal(0.0000115,tb.uniform(0.000002))
        self.delta_alpha = ureal(0.0,tb.uniform(0.000001),50)
        
        self.theta_1 = ureal(-0.1,0.2)
        self.theta_2 = ureal(0.0,tb.arcsine(0.5))
        self.delta_theta = ureal(0.0,tb.uniform(0.05),2)

        self.Ls = ureal(5e7,25,18)        

    def test(self):
        TOL = 1E-10
        
        # Note values from the guide are only approximate
        d = self.d_1 + self.d_2 + self.d_3
        u = uncertainty(d)
        df = dof(d)
        equivalent(u,9.68194195397,TOL)
        equivalent(df,25.4472507774,TOL)

        theta = self.theta_1 + self.theta_2 
        v = value(theta)
        u = uncertainty(theta)
        df = dof(theta)
        equivalent(v,-0.1,TOL)
        equivalent(u,0.406201920232,TOL)
        self.assertTrue( math.isinf(df) )

        x1 = self.Ls * self.delta_alpha * theta       
        u = uncertainty(x1)
        equivalent(u,2.88675134595,TOL)

        x2 = self.Ls * self.alpha_s * self.delta_theta        
        u = uncertainty(x2)
        equivalent(u,16.5988202392,TOL)

        y = self.Ls + d - (x1 + x2)
        u = uncertainty(y)
        df = dof(y)
        equivalent(u,31.6637674111,TOL)
        equivalent(df,16.7521475092,TOL)

#----------------------------------------------------------------------------
class GuideExampleH1SIUnits(unittest.TestCase):
    # Check that the use of very small uncertainty values
    # does not pose a problem.
    
    def setUp(self):
        self.d_1 = ureal(0.0,5.8E-9,24)
        self.d_2 = ureal(0.0,3.9E-9,5)
        self.d_3 = ureal(0.0,6.7E-9,8)

        self.alpha_s = ureal(11.5E-6,tb.uniform(2E-6))
        self.delta_alpha = ureal(0.0,tb.uniform(1E-6),50)
        
        self.theta_1 = ureal(-0.1,0.2)
        self.theta_2 = ureal(0.0,tb.arcsine(0.5))
        self.delta_theta = ureal(0.0,tb.uniform(0.05),2)

        self.Ls = ureal(5e-2,25E-9,18)        

    def test(self):
        TOL = 1E-10
        
        # Note values from the guide are only approximate
        d = self.d_1 + self.d_2 + self.d_3
        u = uncertainty(d)
        df = dof(d)
        equivalent(u,9.68194195397E-9,TOL)
        equivalent(df,25.4472507774,TOL)

        theta = self.theta_1 + self.theta_2 
        v = value(theta)
        u = uncertainty(theta)
        df = dof(theta)
        equivalent(v,-0.1,TOL)
        equivalent(u,0.406201920232,TOL)
        self.assertTrue( math.isinf(df) )

        x1 = self.Ls * self.delta_alpha * theta       
        u = uncertainty(x1)
        equivalent(u,2.88675134595E-9,TOL)

        x2 = self.Ls * self.alpha_s * self.delta_theta        
        u = uncertainty(x2)
        equivalent(u,16.5988202392E-9,TOL)

        y = self.Ls + d - (x1 + x2)
        u = uncertainty(y)
        df = dof(y)
        equivalent(u,31.6637674111E-9,TOL)
        equivalent(df,16.7521475092,TOL)

#-----------------------------------------------------
class GuideExampleH2(unittest.TestCase):

    def setUp(self):
        self.v = ureal(4.999,0.0032,independent=False)
        self.i = ureal(0.019661,0.0000095,independent=False)
        self.phi = ureal(1.04446,0.00075,independent=False)

        set_correlation(-0.36,self.v,self.i)
        set_correlation(0.86,self.v,self.phi)
        set_correlation(-0.65,self.i,self.phi)

    def test_no_dof(self):
        TOL = 1E-10

        v = ureal(4.999,0.0032,independent=False)
        i = ureal(0.019661,0.0000095,independent=False)
        phi = ureal(1.04446,0.00075,independent=False)

        set_correlation(-0.36,v,i)
        set_correlation(0.86,v,phi)
        set_correlation(-0.65,i,phi)

        r = v * cos(phi)/ i
        x = v * sin(phi)/ i
        z = v / i

        equivalent( uncertainty(r),0.0699787279884,TOL)
        equivalent( uncertainty(x),0.295716826846,TOL)
        equivalent( uncertainty(z),0.236602971835,TOL)

        equivalent( math.sqrt(welch_satterthwaite(r)[0]),uncertainty(r),TOL)
        equivalent( math.sqrt(welch_satterthwaite(x)[0]),uncertainty(x),TOL)
        equivalent( math.sqrt(welch_satterthwaite(z)[0]),uncertainty(z),TOL)

        equivalent( get_correlation(r,x),-0.591484610819,TOL)
        equivalent( get_correlation(x,z),0.992797472722,TOL)
        equivalent( get_correlation(r,z),-0.490623905441,TOL)

    def test_with_dof(self):
        TOL = 1E-10

        # The old way! 
        # NB, I think there are now multiple instances of this test 
        # v = ureal(4.999,0.0032,5,independent=False)
        # i = ureal(0.019661,0.0000095,5,independent=False)
        # phi = ureal(1.04446,0.00075,5,independent=False)
        
        v,i,phi = multiple_ureal([4.999,0.019661,1.04446],[0.0032,0.0000095,0.00075],5)

        set_correlation(-0.36,v,i)
        set_correlation(0.86,v,phi)
        set_correlation(-0.65,i,phi)

        r = v * cos(phi)/ i
        x = v * sin(phi)/ i
        z = v / i

        # The presence of finite DoF should not alter previous results
        equivalent( uncertainty(r),0.0699787279884,TOL)
        equivalent( uncertainty(x),0.295716826846,TOL)
        equivalent( uncertainty(z),0.236602971835,TOL)

        equivalent( math.sqrt(welch_satterthwaite(r)[0]),uncertainty(r),TOL)
        equivalent( math.sqrt(welch_satterthwaite(x)[0]),uncertainty(x),TOL)
        equivalent( math.sqrt(welch_satterthwaite(z)[0]),uncertainty(z),TOL)

        equivalent( get_correlation(r,x),-0.591484610819,TOL)
        equivalent( get_correlation(x,z),0.992797472722,TOL)
        equivalent( get_correlation(r,z),-0.490623905441,TOL)

        equivalent( dof(r),5,TOL)
        equivalent( dof(x),5,TOL)
        equivalent( dof(z),5,TOL)
        
        # Add another influence and it breaks again, but earlier results OK
        f = ureal(1,0.0032,5,independent=False)
        self.assertRaises(RuntimeError,set_correlation,0.3,f,i)
        
        equivalent( dof(r),5,TOL)
        equivalent( dof(x),5,TOL)
        equivalent( dof(z),5,TOL)

    def test_with_dof_2(self):
        """
        Same test as above, but now ensure that the 
        ensemble calculation can deal with influences 
        that are not in sequence
        """
        TOL = 1E-10

        # The dummy variables mean that those included 
        # in the ensemble will not be a consecutive 
        # series of IDs, provided we include the 
        # dummies in the equations, as is done below
        # with zero weight.
        v = ureal(4.999,0.0032,5,independent=False)
        dummy_1 = ureal(1,1)
        i = ureal(0.019661,0.0000095,5,independent=False)
        dummy_2 = ureal(1,1)
        phi = ureal(1.04446,0.00075,5,independent=False)
        dummy_3 = ureal(1,1)

        real_ensemble( [v,i,phi],5 )

        set_correlation(-0.36,v,i)
        set_correlation(0.86,v,phi)
        set_correlation(-0.65,i,phi)

        r = v * cos(phi)/ i + (0*dummy_1)
        x = v * sin(phi)/ i + (0*dummy_2)
        z = v / i + (0*dummy_3)

        # The presence of finite DoF should not alter previous results
        equivalent( uncertainty(r),0.0699787279884,TOL)
        equivalent( uncertainty(x),0.295716826846,TOL)
        equivalent( uncertainty(z),0.236602971835,TOL)

        equivalent( math.sqrt(welch_satterthwaite(r)[0]),uncertainty(r),TOL)
        equivalent( math.sqrt(welch_satterthwaite(x)[0]),uncertainty(x),TOL)
        equivalent( math.sqrt(welch_satterthwaite(z)[0]),uncertainty(z),TOL)

        equivalent( get_correlation(r,x),-0.591484610819,TOL)
        equivalent( get_correlation(x,z),0.992797472722,TOL)
        equivalent( get_correlation(r,z),-0.490623905441,TOL)

        equivalent( dof(r),5,TOL)
        equivalent( dof(x),5,TOL)
        equivalent( dof(z),5,TOL) 
        
    def test_with_dof_multiple_ureal(self):
        TOL = 1E-10

        x = [4.999,0.019661,1.04446]
        u = [0.0032,0.0000095,0.00075]
        labels = ['v','i','phi']

        v,i,phi = multiple_ureal(x,u,5,labels)
        
        set_correlation(-0.36,v,i)
        set_correlation(0.86,v,phi)
        set_correlation(-0.65,i,phi)

        r = v * cos(phi)/ i
        x = v * sin(phi)/ i
        z = v / i

        # The presence of finite DoF should not alter previous results
        equivalent( uncertainty(r),0.0699787279884,TOL)
        equivalent( uncertainty(x),0.295716826846,TOL)
        equivalent( uncertainty(z),0.236602971835,TOL)

        equivalent( math.sqrt(welch_satterthwaite(r)[0]),uncertainty(r),TOL)
        equivalent( math.sqrt(welch_satterthwaite(x)[0]),uncertainty(x),TOL)
        equivalent( math.sqrt(welch_satterthwaite(z)[0]),uncertainty(z),TOL)

        equivalent( get_correlation(r,x),-0.591484610819,TOL)
        equivalent( get_correlation(x,z),0.992797472722,TOL)
        equivalent( get_correlation(r,z),-0.490623905441,TOL)

        # Dof calculation should be legal
        equivalent( dof(r),5,TOL)
        equivalent( dof(x),5,TOL)
        equivalent( dof(z),5,TOL)
        
        # Add another influence and it breaks again, but earlier results OK
        f = ureal(1,0.0032,5,independent=False)
        self.assertRaises(RuntimeError,set_correlation,0.3,f,i)
        
        equivalent( dof(r),5,TOL)
        equivalent( dof(x),5,TOL)
        equivalent( dof(z),5,TOL)

        
    def test_with_infinity_1(self):
        """
        Influences may be correlated when they have
        infinite degrees of freedom
        
        """
        df_x3 = 5        
        x1 = ureal(1,1,independent=False)
        x2 = ureal(1,1,independent=False)
        x3 = ureal(1,1,df_x3)
        
        set_correlation(0.1,x1,x2)
        
        y = x1 + x2 + x3

        den = component(y,x3)**4/df_x3
        
        df = dof(y)
        nu_eff = variance(y)**2 / den
        
        self.assertTrue( equivalent(nu_eff,df))

    def test_with_infinity_2(self):
        """
        Complex components may be correlated when they have
        infinite degrees of freedom
        
        """
        df_x1, df_x2 = 5,6,
        r = 0.1
        x1 = ureal(1,1,df_x1)
        x2 = ureal(1,1,df_x2)
        z1 = ucomplex(1,[1,r,r,1])

        m = magnitude(z1)
        
        y = x1 + x2 + m
        
        den = sum([
                component(y,x1)**4/df_x1,
                component(y,x2)**4/df_x2,
            ])
        nu_eff = variance(y)**2 / den
        df = dof(y)
        
        self.assertTrue( equivalent(nu_eff,df))
        
#============================================================================
if(__name__== '__main__'):

    unittest.main()    # Runs all test methods starting with 'test'


"""
Functions
---------    

    Functions for storing and retrieving archive files using Python pickle format are
    
        * :func:`dump`
        * :func:`load`

    Functions for storing and retrieving pickled archive strings are

        * :func:`dumps`
        * :func:`loads`

    Functions for storing and retrieving archive files using JSON format are

        * :func:`dump_json`
        * :func:`load_json`

    Functions for storing and retrieving an archive as a JSON-formatted string are
    
        * :func:`dumps_json`
        * :func:`loads_json`

Module contents
---------------

"""
import json
try:
    import cPickle as pickle  # Python 2
    PY2 = True
except ImportError:
    import pickle
    PY2 = False

from GTC import context
from GTC.archive import Archive
from GTC.json_format import (
    JSONArchiveEncoder,
    json_to_archive,
)

__all__ = (
    'Archive',
    'load',
    'dump',
    'dumps',
    'loads',
    'dump_json',
    'load_json',
    'dumps_json',
    'loads_json',
)

#------------------------------------------------------------------     
def dump(file,ar):
    """Save an archive in a file

    :arg file:  a file object opened in binary mode (with 'wb')
                
    :arg ar: an :class:`Archive` object
      
    Several archives can be saved in the same file 
    by repeated use of this function.
    
    """
    ar._freeze()

# About pickle protocols (from 3.8 docs, abridged)
# 0 is the original 'human-readable' protocol and is backwards compatible with earlier versions of Python.
# 1 is an old binary format which is also compatible with earlier versions of Python.
# 2 was introduced in Python 2.3. It provides much more efficient pickling of new-style classes. 
# 3 was added in Python 3.0. It has explicit support for bytes objects and cannot be unpickled by Python 2.x. 
#    This was the default protocol in Python 3.0 - 3.7.
# 4 was added in Python 3.4. It adds support for very large objects, pickling more kinds of objects, and some data format optimizations. 
#    It is the default protocol starting with Python 3.8. 
# 5 was added in Python 3.8. It adds support for out-of-band data and speedup for in-band data. 

    pickle.dump(ar,file,protocol=2)     # Change to 3 when GTC no longer supports Python 2.7

#------------------------------------------------------------------     
def load(file):
    """Load an archive from a file

    :arg file:  a file object opened in binary mode (with 'rb')

    Several archives can be extracted from 
    the same file by repeatedly calling this function.
    
    """
    ar = pickle.load(file)
    ar.context = context._context
    ar._thaw()
    
    return ar

#------------------------------------------------------------------     
def dumps(ar,protocol=pickle.HIGHEST_PROTOCOL):
    """
    Save an archive pickled in a string  

    :arg ar: an :class:`Archive` object
    :arg protocol: encoding type 

    Possible values for :ref:`protocol <pickle-protocols>` are described in the
    Python documentation for the :mod:`pickle` module.

    ``protocol=0`` creates an ASCII string, but note
    that many (special) linefeed characters are embedded.
    
    """
    # Can save one of these strings in a single binary file,
    # using write(), when protocol=pickle.HIGHEST_PROTOCOL is used. 
    # A corresponding read() is required to extract the string. 
    # Alternatively, when protocol=0 is used a text file can be 
    # used, but again write() and read() have to be used, 
    # because otherwise the embedded `\n` characters are 
    # interpreted incorrectly.
    
    ar._freeze()
    s = pickle.dumps(ar,protocol)
    
    return s
    
#------------------------------------------------------------------     
def loads(s):
    """
    Return an archive object from a pickled string 

    :arg s: a string created by :func:`dumps`
    
    """
    ar = pickle.loads(s)
    ar.context = context._context
    ar._thaw()
    
    return ar

#------------------------------------------------------------------     
def dumps_json(ar,**kw):
    """
    Convert an archive to a JSON string  

    :arg ar: an :class:`Archive` object
    
    Keyword arguments will be passed to :func:`json.dumps()`

    .. versionadded:: 1.3.0
    """
    ar._freeze()
    s = json.dumps(ar, cls=JSONArchiveEncoder,**kw )
    
    return s

#------------------------------------------------------------------     
def loads_json(s,**kw):
    """
    Return an archive object by converting a JSON string  

    :arg s: a string created by :func:`dumps_json`
    
    Keyword arguments will be passed to :func:`json.loads()`
    
    .. versionadded:: 1.3.0
    """
    ar = json.loads(s,object_hook=json_to_archive,**kw)    
    ar.context = context._context
    ar._thaw()
    
    return ar
    
#------------------------------------------------------------------     
def dump_json(file,ar,**kw):
    """Save an archive in a file in JSON format

    :arg file:  a file object opened in text mode (with 'w')
                
    :arg ar: an :class:`Archive` object
      
    Keyword arguments will be passed to :func:`json.dump()`

    Only one archive can be saved in a file.
    
    .. versionadded:: 1.3.0
    """
    ar._freeze()
    s = json.dump(ar, file, cls=JSONArchiveEncoder, **kw )
    
    return s

#------------------------------------------------------------------     
def load_json(file,**kw):
    """Load an archive from a file

    :arg file: a file created by :func:`dump_json`
    
    Keyword arguments will be passed to :func:`json.load()`

    .. versionadded:: 1.3.0
    """
    ar = json.load(file, object_hook=json_to_archive,**kw)    
    ar.context = context._context
    ar._thaw()
    
    return ar
    
#============================================================================    
if __name__ == "__main__":
    import doctest
    from GTC import *  
    doctest.testmod( optionflags=doctest.NORMALIZE_WHITESPACE )



import re
import unittest
import numpy

from GTC import *
from GTC import context 
from GTC.vector import *
from GTC.nodes import *
from GTC.reporting import u_component
from GTC.lib import (
    UncertainReal,
    welch_satterthwaite,
    std_variance_real,
    set_correlation_real,
    get_correlation_real,
    real_ensemble,
    append_real_ensemble,
    _is_uncertain_real_constant
)

from testing_tools import *

TOL = 1E-13 
LOG10_E = math.log10(math.e)

# numpy warning are raised as errors
numpy.seterr(all='raise')

#----------------------------------------------------------------------------
class SimpleAttributesAndFunctions(unittest.TestCase):

    # An UncertainReal object has properties `x`, `u`, `v` and `df`.
    # There are corresponding functions: `value()`, `uncertainty()`, 
    # `variance()` and `dof()` 
    
    # An UncertainReal object also has `real`, `imag` and `conjugate`
    # properties.
    
    # An UncertainReal object may be implicitly converted to a 
    # float, a complex or a Boolean. 
    
    # The unary positive operation applied to an UncertainReal object 
    # creates a clone object.
    
    # The absolute value returns a float. 
    
    # The truth value of an uncertain number is equal to 

    #------------------------------------------------------------------------
    def setUp(self):
        self.x = x = 10.2
        self.u = u = 3.2
        self.df = df = 5
        self.label = label = 'me_too'
        
        self.un = ureal(x,u,df,label)
        self.const = constant(x,label)    
            
    def test_value(self):
        self.assertTrue( equivalent(self.x,value(self.un) ) )
        self.assertTrue( equivalent(self.x,value(self.const) ) )
        self.assertTrue( equivalent(self.x,self.un.x ) )
        self.assertTrue( equivalent(self.x,self.const.x ) )
      
    def test_uncertainty(self):
        self.assertTrue( equivalent(self.u,uncertainty(self.un) ) )
        self.assertTrue( equivalent(0,uncertainty(self.const) ) )
        self.assertTrue( equivalent(self.u,self.un.u ) )
        self.assertTrue( equivalent(0,self.const.u ) )
      
    def test_variance(self):
        self.assertTrue( equivalent(self.u**2,variance(self.un) ) )
        self.assertTrue( equivalent(0,variance(self.const) ) )
        self.assertTrue( equivalent(self.u**2,self.un.v ) )
        self.assertTrue( equivalent(0,self.const.v ) )
      
    def test_dof(self): 
        self.assertTrue( equivalent(self.df,dof(self.un) ) )
        self.assertTrue( equivalent(inf,dof(self.const) ) )
        self.assertTrue( equivalent(self.df,self.un.df ) )
        self.assertTrue( equivalent(inf,self.const.df ) )
        self.assertTrue( math.isinf( self.const.df ) )
 
    def test_conjugate(self):
        self.assertTrue( self.un.conjugate() is self.un )
        
    def test_real_imaginary(self):
        self.assertTrue( self.un.real == self.x )
        self.assertTrue( self.un.imag == 0 )
        
    def test_conversions(self):
        self.assertTrue( equivalent( value(self.un), self.x )  )
        z = value( self.un )
        self.assertTrue( equivalent( z.real, self.x ) )
        self.assertTrue( z.imag == 0 )
        self.assertTrue( bool(z.imag) == False )
        self.assertTrue( bool(z.real) == True )
        
    def test_positive_unary(self):
        un_plus = +self.un
        self.assertTrue(un_plus is not self.un)
        self.assertTrue( equivalent(self.x,value(un_plus) ) )
        self.assertTrue( equivalent(self.u,uncertainty(un_plus) ) )
        self.assertTrue( equivalent(self.df,dof(un_plus) ) )

    def test_absolute(self):
        un_minus = -self.un
        self.assertTrue( isinstance( abs(un_minus), float ) )
        self.assertTrue( equivalent(self.un.x,abs(un_minus) ) )

    def test_non_zero(self):
        x = ureal(1,1)
        self.assertTrue( bool(x) is True )
        x = ureal(0,1)
        self.assertTrue( bool(x) is not True )
        
    def test_is_constant(self): 
        x = constant(1.1)
        self.assertTrue( _is_uncertain_real_constant(x) )
        x = ureal(1,1)
        self.assertTrue( _is_uncertain_real_constant(x) is not True )
        x= 3
        self.assertRaises( TypeError, _is_uncertain_real_constant, x )
        
#----------------------------------------------------------------------------
class ArithmeticTestsReal(unittest.TestCase):

    # 
    
    #------------------------------------------------------------------------
    def setUp(self):
        self.x = ureal(1.,2.)
        self.w = ureal(2.,2.)
        self.z = ureal(3.,3.)
        
    #------------------------------------------------------------------------
    def testAddition(self):
        # Cases to consider: un-un, un-numb, numb-un
        # Need to check u_component in each case.
        # Adding zero is a special case
        
        #---------------------------------------------------------
        # += operator is not defined, but we check it
        x = self.x
        x += self.z
        equivalent(value(x),value(self.x)+value(self.z),TOL)
        
        equivalent(
            uncertainty(x)
        ,   math.sqrt( uncertainty(self.x)**2 + uncertainty(self.z)**2 )
        ,   TOL
        )
        self.assertTrue( math.isinf( dof(x) ) )

        #---------------------------------------------------------
        # Adding a constant on the right
        x = x + 10
        equivalent(value(x),value(self.x)+value(self.z)+10,TOL)
        
        equivalent(
            uncertainty(x)
        ,   math.sqrt(uncertainty(self.x)**2+uncertainty(self.z)**2)
        ,   TOL
        )

        # component of uncertainty
        x1 = result( self.x + self.w )
        x2 = x1 + 4
        equivalent(
            u_component(x2,x1)
        ,   uncertainty(x1)
        ,   TOL
        )
        
        #---------------------------------------------------------
        # Adding a constant on the left
        x = self.x + self.z
        x = 10 + x
        equivalent(value(x),value(self.x)+value(self.z)+10,TOL)
        
        equivalent(
            uncertainty(x)
        ,   math.sqrt(uncertainty(self.x)**2+uncertainty(self.z)**2)
        ,   TOL
        )
        
        # component of uncertainty
        x1 = result( self.x + self.w )
        x2 = 3 + x1
        equivalent(
            u_component(x2,x1)
        ,   uncertainty(x1)
        ,   TOL
        )

        #---------------------------------------------------------
        # regular addition
        y = self.x + self.w
        equivalent(value(y),3.0,TOL)
        
        equivalent(uncertainty(y),2.0 * math.sqrt(2.0),TOL)
        self.assertTrue( math.isinf( dof(y) ) )

        # component of uncertainty
        x1 = result( y )
        x2 = x1 + self.z
        
        equivalent(value(x2),6.0,TOL)
        equivalent(
            u_component(x2,x1)
        ,   uncertainty(x1)
        ,   TOL
        )

        #---------------------------------------------------------
        # trivial addition
        oldy = y
        y += 0
        self.assertTrue(oldy is y)
        y = y + 0
        self.assertTrue(oldy is y)
        y = 0 + y
        self.assertTrue(oldy is y)
        
        newy = y + 0j
        self.assertFalse( newy is y )
        self.assertTrue( equivalent(value(newy.real),value(y),TOL) )
        self.assertTrue( equivalent(value(newy.imag),0.0,TOL) )
        self.assertTrue( equivalent(uncertainty(newy)[0],uncertainty(y),TOL) )
        self.assertTrue( equivalent(uncertainty(newy)[1],0.0,TOL) )

        newy = 0j + y 
        self.assertFalse( newy is y )
        self.assertTrue( equivalent(value(newy.real),value(y),TOL) )
        self.assertTrue( equivalent(value(newy.imag),0.0,TOL) )
        self.assertTrue( equivalent(uncertainty(newy)[0],uncertainty(y),TOL) )
        self.assertTrue( equivalent(uncertainty(newy)[1],0.0,TOL) )

        #---------------------------------------------------------
        # Adding a complex constant on the left
        x = self.x 
        z = -10+6j
        y = z + x
        
        re = y.real 
        im = y.imag 
        
        equivalent(
            value(y.real),
            value(x)+z.real,
            TOL
        )
  
        equivalent(
            value(y.imag),
            z.imag,
            TOL
        )
  
        equivalent(
            uncertainty(y.real)
        ,   x.real.u
        ,   TOL
        )

        equivalent(
            uncertainty(y.imag)
        ,   0.0
        ,   TOL
        )  

        #---------------------------------------------------------
        # Adding a complex constant on the right
        x = self.x 
        z = 10-3j
        y = x + z
        
        re = y.real 
        im = y.imag 
        
        equivalent(
            value(y.real),
            value(x)+z.real,
            TOL
        )
  
        equivalent(
            value(y.imag),
            z.imag,
            TOL
        )
  
        equivalent(
            uncertainty(y.real)
        ,   x.real.u
        ,   TOL
        )

        equivalent(
            uncertainty(y.imag)
        ,   0.0
        ,   TOL
        )
                
    #------------------------------------------------------------------------
    def testSubtraction(self):
        # Cases to consider: un-un, un-numb, numb-un
        # Need to check u_component in each case.

        #---------------------------------------------------------
        # -= operator is not defined, but we check it
        x = self.x
        x -= self.z
        equivalent(value(x),value(self.x)-value(self.z),TOL)
        
        equivalent(
            uncertainty(x)
        ,   math.sqrt(uncertainty(self.x)**2+uncertainty(self.z)**2)
        ,   TOL
        )
        self.assertTrue( math.isinf( dof(x) ) )

        #---------------------------------------------------------
        # Subtracting a constant
        x = x - 10
        equivalent(value(x),value(self.x)-value(self.z)-10,TOL)

        equivalent(
            uncertainty(x)
        ,   math.sqrt(uncertainty(self.x)**2+uncertainty(self.z)**2)
        ,   TOL
        )
        
        # component of uncertainty
        x1 = result( self.x + self.w )
        x2 = x1 - 10
        equivalent(
            u_component(x2,x1)
        ,   uncertainty(x1)
        ,   TOL
        )

        #---------------------------------------------------------
        # lhs is not an UN
        x = 10 - x
        equivalent(value(x),-value(self.x)+value(self.z)+20,TOL)

        equivalent(
            uncertainty(x)
        ,   math.sqrt(uncertainty(self.x)**2+uncertainty(self.z)**2)
        ,   TOL
        )

        # component of uncertainty
        x1 = result( self.x + self.w )
        x2 = 10 - x1
        equivalent(
            u_component(x2,x1)
        ,   -uncertainty(x1)
        ,   TOL
        )

        #---------------------------------------------------------
        # regular subtraction        
        y = self.x - self.w
        equivalent(value(y),value(self.x)-value(self.w))
        equivalent(uncertainty(y),2.0 * math.sqrt(2.0))

        # component of uncertainty
        x1 = result( y )
        x2 = x1 - self.z
        equivalent(
            u_component(x2,x1)
        ,   uncertainty(x1)
        ,   TOL
        )
        
        #---------------------------------------------------------
        # trivial subtraction
        oldy = y
        y -= 0
        self.assertTrue(oldy is y)
        y = y - 0
        self.assertTrue(oldy is y)
        y = 0 - y
        self.assertTrue(oldy is not y)
        equivalent(value(y),-value(oldy))

        newy = y - 0j
        self.assertFalse( newy is y )
        self.assertTrue( equivalent(value(newy.real),value(y),TOL) )
        self.assertTrue( equivalent(value(newy.imag),0.0,TOL) )
        self.assertTrue( equivalent(uncertainty(newy)[0],uncertainty(y),TOL) )
        self.assertTrue( equivalent(uncertainty(newy)[1],0.0,TOL) )

        newy = 0j - y 
        self.assertFalse( newy is y )
        self.assertTrue( equivalent(value(newy.real),-value(y),TOL) )
        self.assertTrue( equivalent(value(newy.imag),0.0,TOL) )
        self.assertTrue( equivalent(uncertainty(newy)[0],uncertainty(y),TOL) )
        self.assertTrue( equivalent(uncertainty(newy)[1],0.0,TOL) )
        
        #---------------------------------------------------------
        # Subtracting a complex constant on the left
        x = self.x 
        z = -10+6j
        y = z - x
        
        re = y.real 
        im = y.imag 
        
        equivalent(
            value(y.real),
            z.real-value(x),
            TOL
        )
  
        equivalent(
            value(y.imag),
            z.imag,
            TOL
        )
  
        equivalent(
            uncertainty(y.real)
        ,   x.real.u
        ,   TOL
        )

        equivalent(
            uncertainty(y.imag)
        ,   0.0
        ,   TOL
        )  

        #---------------------------------------------------------
        # Subtracting a complex constant on the right
        x = self.x 
        z = 3.2+6.1j
        y = x - z
        
        re = y.real 
        im = y.imag 
        
        equivalent(
            value(y.real),
            value(x)-z.real,
            TOL
        )
  
        equivalent(
            value(y.imag),
            -z.imag,
            TOL
        )
  
        equivalent(
            uncertainty(y.real)
        ,   x.real.u
        ,   TOL
        )

        equivalent(
            uncertainty(y.imag)
        ,   0.0
        ,   TOL
        )  
        
    #------------------------------------------------------------------------
    def testMultiplication(self):
        # Cases to consider: un-un, un-numb, numb-un
        # Need to check u_component in each case.

        #---------------------------------------------------------
        # *= operator is not defined, but we check it
        x = self.x
        x *= self.z
        equivalent(value(x),value(self.x)*value(self.z),TOL)

        equivalent(
            uncertainty(x)
        ,   math.sqrt(
                (value(self.z) * uncertainty(self.x))**2
            +   (value(self.x) * uncertainty(self.z))**2
            )
        ,   TOL
        )
        self.assertTrue( math.isinf( dof(x) ) )

        #---------------------------------------------------------
        # multiplying by a constant
        x = x * 10
        equivalent(value(x),(value(self.x)*value(self.z))*10,TOL)

        equivalent(
            uncertainty(x)
        ,   math.sqrt(
                (10*value(self.z) * uncertainty(self.x))**2
            +   (10*value(self.x) * uncertainty(self.z))**2
            )
        ,   TOL
        )
        
        # component of uncertainty
        x1 = result( self.x + self.w )
        x2 = x1 * 5
        equivalent(
            u_component(x2,x1)
        ,   5 * uncertainty(x1)
        ,   TOL
        )

        #---------------------------------------------------------
        # lhs is not an UN
        x = 10 * x
        equivalent(value(x),(value(self.x)*value(self.z))*100,TOL)
        equivalent(
            uncertainty(x)
        ,   math.sqrt(
                (100*value(self.z) * uncertainty(self.x))**2
            +   (100*value(self.x) * uncertainty(self.z))**2
            )
        ,   TOL
        )

        # component of uncertainty
        x1 = result( self.x + self.w )
        x2 = 3 * x1
        equivalent(
            u_component(x2,x1)
        ,   3 * uncertainty(x1)
        ,   TOL
        )

        #---------------------------------------------------------
        # regular multiplication
        y = self.z * self.w
        equivalent(value(y),6.0)
        equivalent(uncertainty(y),6.0 * math.sqrt(2.0))

        # component of uncertainty
        x1 = result( self.x * self.w )
        x2 = self.z * x1
        equivalent(
            u_component(x2,x1)
        ,   value(self.z) * uncertainty(x1)
        ,   TOL
        )

        #---------------------------------------------------------
        # trivial multiplication
        oldy = y
        y *= 1
        self.assertTrue(oldy is y)
        y = y * 1
        self.assertTrue(oldy is y)
        y = 1.0 * y
        self.assertTrue(oldy is y)

        newy = y*(1.0+0j)
        self.assertFalse( newy is y )
        self.assertTrue( equivalent(value(newy.real),value(y),TOL) )
        self.assertTrue( equivalent(value(newy.imag),0.0,TOL) )
        self.assertTrue( equivalent(uncertainty(newy)[0],uncertainty(y),TOL) )
        self.assertTrue( equivalent(uncertainty(newy)[1],0.0,TOL) )

        newy = (1.0+0j)*y
        self.assertFalse( newy is y )
        self.assertTrue( equivalent(value(newy.real),value(y),TOL) )
        self.assertTrue( equivalent(value(newy.imag),0.0,TOL) )
        self.assertTrue( equivalent(uncertainty(newy)[0],uncertainty(y),TOL) )
        self.assertTrue( equivalent(uncertainty(newy)[1],0.0,TOL) )
        
        #---------------------------------------------------------
        # Multiplying a complex constant on the left
        x = self.x 
        z = 10+4j
        y = z*x
        
        re = y.real 
        im = y.imag 
        
        equivalent(
            value(y.real),
            (z*value(x)).real,
            TOL
        )
  
        equivalent(
            value(y.imag),
            (z*value(x)).imag,
            TOL
        )
  
        equivalent(
            uncertainty(y.real)
        ,   z.real * x.u
        ,   TOL
        )

        equivalent(
            uncertainty(y.imag)
        ,   z.imag * x.u
        ,   TOL
        )  

        #---------------------------------------------------------
        # Multiplying a complex constant on the right
        x = self.x 
        z = 3.2+6.1j
        y = x * z
        
        equivalent(
            value(y.real),
            (z*value(x)).real,
            TOL
        )
    
        equivalent(
            value(y.imag),
            (value(x)*z).imag,
            TOL
        )
  
        equivalent(
            uncertainty(y.real)
        ,   x.u * z.real
        ,   TOL
        )

        equivalent(
            uncertainty(y.imag)
        ,   x.u * z.imag
        ,   TOL
        )  
    #------------------------------------------------------------------------
    def testDivision(self):
        # Cases to consider: un-un, un-numb, numb-un
        # Need to check u_component in each case.

        #---------------------------------------------------------
        # /= operator is not defined, but we check it
        x = self.x
        x /= self.z
        equivalent(value(x),value(self.x)/value(self.z),TOL)
        equivalent(
            uncertainty(x)/value(x)
        ,   math.sqrt(
                (uncertainty(self.x)/value(self.x))**2
            +   (uncertainty(self.z)/value(self.z))**2
            )
        ,   TOL
        )
        self.assertTrue( math.isinf(dof(x)) )

        #---------------------------------------------------------
        # regular division
        y = self.z / self.w
        equivalent(value(y),1.5)
        equivalent(uncertainty(y),1.5 * math.sqrt(2.0))

        # component of uncertainty
        x1 = result( self.x / self.w )
        x2 = x1 / self.x
        equivalent(
            u_component(x2,x1)
        ,   value(x1)/value(x2) * uncertainty(x1)
        ,   TOL
        )
        #---------------------------------------------------------
        # Divide by a constant
        y = self.w / 5.0
        equivalent(value(y),value(self.w)/5.0,TOL)
        equivalent(
            uncertainty(y)
        ,   uncertainty(self.w) / 5.0
        ,   TOL
        )
        equivalent(
            u_component(y,self.w)
        ,   uncertainty(self.w) / 5.0
        ,   TOL
        )
        
        # component of uncertainty
        x1 = result( self.x / self.w )
        x2 = x1 / 5.0
        equivalent(
            u_component(x2,x1)
        ,   uncertainty(x1) / 5
        ,   TOL
        )
        #---------------------------------------------------------
        # lhs is not UN
        y = 4.0 / self.w
        equivalent(value(y),4.0 / value(self.w),TOL)
        equivalent(
            u_component(y,self.w)
        ,   -4.0 / value(self.w)**2 * uncertainty(self.w) 
        ,   TOL
        )

        # component of uncertainty
        x1 = result( self.x / self.w )
        x2 = 4.0 / x1
        equivalent(
            u_component(x2,x1)
        ,   -4.0 * uncertainty(x1) / value(x1)**2
        ,   TOL
        )
        
        #---------------------------------------------------------
        # trivial division
        y = self.w / 1.0
        self.assertTrue( y is self.w )
        y = 1.0 / self.w
        self.assertTrue( y is not self.w )
        
        self.assertRaises(ZeroDivisionError,UncertainReal.__div__, self.z, 0)

        newy = y/(1.0+0j)
        self.assertFalse( newy is y )
        self.assertTrue( equivalent(value(newy.real),value(y),TOL) )
        self.assertTrue( equivalent(value(newy.imag),0.0,TOL) )
        self.assertTrue( equivalent(uncertainty(newy)[0],uncertainty(y),TOL) )
        self.assertTrue( equivalent(uncertainty(newy)[1],0.0,TOL) )

        newy = (1.0+0j)/y
        self.assertFalse( newy is y )
        self.assertTrue( equivalent(value(newy.real),1./value(y),TOL) )
        self.assertTrue( equivalent(value(newy.imag),0.0,TOL) )
        self.assertTrue( 
            equivalent(
                uncertainty(newy)[0]/value(newy.real),
                uncertainty(y)/value(y),TOL
            ) 
        )
        self.assertTrue( equivalent(uncertainty(newy)[1],0.0,TOL) )
        

        #---------------------------------------------------------
        # Dividing a complex constant on the left
        x = self.x 
        z = 10+4j
        y = z/x
        
        re = y.real 
        im = y.imag 
        
        equivalent(
            value(y.real),
            (z/value(x)).real,
            TOL
        )
  
        equivalent(
            value(y.imag),
            (z/value(x)).imag,
            TOL
        )
  
        norm = value(x)**2
        equivalent(
            uncertainty(y.real)
        ,   abs( z.real/norm*x.u )
        ,   TOL
        )

        equivalent(
            uncertainty(y.imag)
        ,   abs( z.imag/norm*x.u )
        ,   TOL
        )  

        #---------------------------------------------------------
        # Dividing by a complex constant on the right
        x = self.x 
        z = 3.2+6.1j
        y = x / z
        
        equivalent(
            value(y.real),
            (value(x)/z).real,
            TOL
        )
    
        equivalent(
            value(y.imag),
            (value(x)/z).imag,
            TOL
        )
  
        norm = abs(z)**2
        equivalent(
            uncertainty(y.real)
        ,   x.u * abs(z.real)/norm
        ,   TOL
        )

        equivalent(
            uncertainty(y.imag)
        ,   x.u * abs(z.imag)/norm
        ,   TOL
        )  
    #------------------------------------------------------------------------
    def testNegation(self):
        x = self.x
        w = self.w
        y = -(x + w)
        equivalent(value(y),-value(value(self.x)+value(self.w)),TOL)
        u_x = u_component(y,self.x)
        equivalent(u_x,-u_component(self.x+self.w,self.x),TOL)

        # component of uncertainty
        x1 = result( self.x / self.w )
        x2 = -x1
        equivalent(
            u_component(x2,x1)
        ,   -1.0 * uncertainty(x1)
        ,   TOL
        )

#----------------------------------------------------------------------------
class TestUncertainReal(unittest.TestCase):
    """
    UncertainReal object special methods.
    Some of these handle mathematical operations and are tested
    elsewhere. Here testing is for repr() and creation 
    
    """
    def test_elementary(self):
        x = ureal(1,1)
        self.assertTrue( x.is_elementary )
        self.assertTrue( not x.is_intermediate )
        self.assertEqual( None, x.label )
        
        y = x + 10
        self.assertTrue( not y.is_elementary )
        self.assertTrue( not y.is_intermediate )
        self.assertEqual( None, y.label )
        
        y = result(x + 10,label='y')
        self.assertTrue( y.is_intermediate )
        self.assertEqual( 'y', y.label )
    
        z = constant(10,label='z')
        self.assertTrue( not z.is_elementary )
        self.assertTrue( not z.is_intermediate )
        self.assertEqual( 'z', z.label )
            
    def test(self):
        """
        The constructor takes arguments: context,x,u_comp,d_comp,i_comp [,node]
        
        A node object is provided when creating an elementary uncertain number 
        or an intermediate result.
        
        """
        x = 1.
        u = 0.1 
        df = 'inf'
        
        un = ureal(x,u)
        s = re.search(r'ureal\((.*),(.*),(.*)\)', repr(un))
        self.assertEqual(s.group(1), repr(x) )
        self.assertEqual(s.group(2), repr(u) )
        self.assertEqual(s.group(3), df )
        
        x = 1.2
        u = 0.2
        x_str= "1.20(20)" # takes account of uncertainty rounding

        un = ureal(x,u)

        rep = "ureal({!r},{!r},inf)".format(x,u)
        self.assertEqual( rep,repr(un) )
        self.assertEqual( x_str,str(un) )

        self.assertEqual( value(un), x)

        # These tests just look at the constructor and by-pass the context. 
        # Here is an uncertain number with no node
        u_comp = Vector()
        u_comp.extend( [(1,u)] )
        d_comp = Vector()
        d_comp.extend( [(2,2*u)] )
        i_comp = Vector()
        un = UncertainReal(x,u_comp,d_comp,i_comp)
        self.assertTrue( un._u_components is u_comp )
        self.assertTrue( un._i_components is i_comp )
        self.assertTrue( un._d_components is d_comp )

        # Here is an uncertain number with a node
        # args: uid,label,u,df [,independent=True]
        label = 'test'
        df = 10
        lf = Leaf(3,label,u,df)
        un = UncertainReal(x,u_comp,d_comp,i_comp,lf)
        self.assertTrue( un._u_components is u_comp )
        self.assertTrue( un._i_components is i_comp )
        self.assertTrue( un._d_components is d_comp )
        self.assertTrue( un._node.label is label )
        self.assertTrue( un._node.independent is True )
        self.assertTrue( un._node.df == df )
        
    def test_check_identity(self):
        """
        The `reporting.is_ureal` function should be able to pick
        an uncertain real number.
        
        """
        x = ureal(1,1)
        self.assertTrue( reporting.is_ureal(x) )

        y = ureal(1,1) * x
        self.assertTrue( reporting.is_ureal(y) )

        z = constant(6)        
        self.assertTrue( reporting.is_ureal(z) )

        z1 = ucomplex(1+3j,1)
        self.assertTrue( reporting.is_ureal( z1.real ) )
        
    def test_strange_rounding_cases(self):
        """
        The __str__ method 
        
        """
        x = UncertainReal._elementary(
            inf,inf,inf,
            None,True
        )
        self.assertEqual( str(x), 'inf(inf)')
        self.assertEqual( repr(x), 'ureal(inf,inf,inf)')

        x = UncertainReal._elementary(
            inf,inf,nan,
            None,True
        )
        self.assertEqual( str(x), 'inf(inf)')
        self.assertEqual( repr(x), 'ureal(inf,inf,nan)')
        
        x = UncertainReal._elementary(
            inf,nan,nan,
            None,True
        )
        self.assertEqual( str(x), 'inf(nan)')
        self.assertEqual( repr(x), 'ureal(inf,nan,nan)')
        
        x = UncertainReal._elementary(
            nan,nan,nan,
            None,True
        )
        self.assertEqual( str(x), 'nan(nan)')
        self.assertEqual( repr(x), 'ureal(nan,nan,nan)')

        x = UncertainReal._elementary(
            1,0.1,nan,
            None,True
        )
        self.assertEqual( str(x), '1.00(10)')
        self.assertEqual( repr(x), 'ureal(1.0,0.1,nan)')

        x = ureal(1.0,0.0)
        self.assertEqual( str(x), '1.000000')   # reverts to default number format
        self.assertEqual( repr(x), 'ureal(1.0,0.0,inf)')

#----------------------------------------------------------------------------
class TestComparisons(unittest.TestCase):

    # UncertainReal defines comparisons wrt the object value 
    
    #------------------------------------------------------------------------
    def test_ureal_comparison(self):
        # All float comparisons should apply
        x_value = -1.2
        x_greater = x_value + .1
        x_less = x_value - .1
        
        x = ureal(x_value,0.5)
        x_g = ureal(x_greater,0.25)
        x_l = ureal(x_less,1)
        x_false = ureal(0,1)

        self.assertTrue(x == x_value)
        self.assertTrue(x <= x_greater)
        self.assertTrue(x >= x_less)
        self.assertTrue(x < x_greater)
        self.assertTrue(x > x_less)
        self.assertTrue(x != x_greater)
        self.assertTrue(x != x_less)

        self.assertTrue(x <= x_g)
        self.assertTrue(x >= x_l)
        self.assertTrue(x < x_g)
        self.assertTrue(x > x_l)
        self.assertTrue(x != x_g)
        self.assertTrue(x != x_l)
        
        # Boolean tests depend on value 
        self.assertTrue( bool(x) )
        self.assertTrue( not bool(x_false) )
        self.assertTrue( x )
        self.assertTrue( not x_false )

        self.assertTrue( equivalent(abs(x_value),abs(x)) )

#----------------------------------------------------------------------------
class TestFunctionsReal(unittest.TestCase):
    
    def setUp(self):
        self.x1 = 1.7
        self.u1 = 0.1
        self.un1 = ureal(self.x1,self.u1) 
        
        self.x2 = 0.5
        self.u2 = 0.4
        self.un2 = ureal(self.x2,self.u2)
        
        self.x3 = -1.0
        self.u3 = 0.4
        self.un3 = ureal(self.x3,self.u3)

        self.x4 = 0.0
        self.u4 = 0.2
        self.un4 = ureal(self.x4,self.u4)
        
        self.x5 = 1.0
        self.u5 = 0.2
        self.un5 = ureal(self.x5,self.u5)

        # Create intermediate values for u_component checking
        self.x_12 = self.x1 * self.x2
        self.u_12 = math.sqrt((self.x1 * self.u2)**2 + (self.x2 * self.u1)**2)

        # By using `result` here the i_components buffer will be used
        # in the intermediate tests. 
        self.un_12 = result( self.un1 * self.un2 ) 
        
        self.un_22 = result( self.un2 / ureal(.3,.1) )

    # This test case is probably unnecessary
    # def test_conjugate(self):
        # """
        # conjugate() is implemented so as to simply return self
        
        # """
        # self.assertTrue( equivalent(self.un1.x, self.un1.conjugate().x) )
        # self.assertTrue( equivalent(self.un1.u, self.un1.conjugate().u) )
        # self.assertTrue( self.un1 is self.un1.conjugate() )
        
        # # 
        # x2 = result( +self.un5 )
        # xc = x2.conjugate() 
        # self.assertTrue(
            # equivalent_matt(
                # xc._i_components,
                # x2._i_components,
                # TOL
        # ))
 
    def test_Pwr(self):
        # ur-ur
        vv = self.x1 ** self.x2
        # The two components of uncertainty are:
        u1 = self.x2 * vv / self.x1 * self.u1
        u2 = vv * math.log( abs( self.x1) ) * self.u2
        y = self.un1 ** self.un2
        v = value(y)
        equivalent( v,  vv, TOL  )
        equivalent( u1, u_component(y,self.un1), TOL )
        equivalent( u2, u_component(y,self.un2), TOL )

        # 0 ** 1 is a tricky one
        y = self.un4 ** self.un5
        v = value(y)
        equivalent( v,  0., TOL  )
        equivalent( self.u4, u_component(y,self.un4), TOL )
        equivalent( 0., u_component(y,self.un5), TOL )

        # y = x ** 1                             
        y = self.un2 ** self.un5
        
        # The two components of uncertainty are:
        u1 = self.x5 * (self.x2 ** (self.x5 - 1)) * self.u2
        u2 = (self.x2 ** self.x5) * math.log( abs( self.x2) ) * self.u5

        v = value(y)
        equivalent( v,  self.x2 ** self.x5 , TOL  )
        equivalent( u1, u_component(y,self.un2), TOL )
        equivalent( u2, u_component(y,self.un5), TOL )

        # x ** number
        number = 2.5
        y = self.un1 ** number
        vv = self.x1 ** number
        v = value(y)
        equivalent( v,  vv, TOL  )
        # The components of uncertainty are:
        u1 = number * vv / self.x1 * self.u1        
        equivalent( u1, u_component(y,self.un1), TOL )

        # number ** x
        number = 2.5
        y = number ** self.un2
        vv = number ** self.x2 
        v = value(y)
        equivalent( v,  vv, TOL  )
        # The components of uncertainty are:
        u2 = vv * math.log( abs( number ) ) * self.u2
        equivalent( u2, u_component(y,self.un2), TOL )
 
        # x ** complex number
        number = .5+1.2j
        y = self.un1 ** number
        vv = self.x1 ** number
        v = value(y)
        equivalent_complex( v,  vv, TOL  )
        # The real/imag components of uncertainty are:
        dy_dx = number * vv / self.x1  
        u_re = abs(dy_dx.real * self.u1)
        u_im = abs(dy_dx.imag * self.u1)
        equivalent( u_im, y.imag.u, TOL )
        equivalent( u_re, y.real.u, TOL )

        # complex number ** x
        number = 2.5-.3j
        y = number ** self.un2
        vv = number ** self.x2 
        v = value(y)
        equivalent_complex( v,  vv, TOL  )
        # The real/imag components of uncertainty are:
        dy_dx = vv * cmath.log( number ) 
        u_re = abs(dy_dx.real * self.u2)
        u_im = abs(dy_dx.imag * self.u2)
        equivalent( u_im, y.imag.u, TOL )
        equivalent( u_re, y.real.u, TOL )
 
        # x ** 1
        y = self.un1 ** 1.0
        self.assertTrue( y is self.un1 )
        
        # x ** 0
        y = self.un1 ** 0.0
        equivalent( 1.0, value(y), TOL  )
        equivalent( 0.0, uncertainty(y), TOL  )

        # intermediate case
        y = self.un_12 ** self.un_22
        
        equivalent(
            value(self.un_22)*value(self.un_12)**(value(self.un_22)-1) * uncertainty(self.un_12),
            u_component(y,self.un_12),
            TOL )
        equivalent(
            math.log( abs(value(self.un_12)) ) * value(y) * uncertainty(self.un_22),
            u_component(y,self.un_22),
            TOL )
          

    def test_magnitude(self):
        """
        magnitude() is implemented as if it were sqrt(x**2)
        
        """
        self.assertRaises(ZeroDivisionError,magnitude,ureal(0,1))

        mag_x3 = magnitude(self.un3)
        equivalent( value(mag_x3), abs( self.x3 ), TOL )

        y = -1 * self.un2
        mag_y = magnitude(y)
        equivalent( value(mag_y), self.x2, TOL )
        equivalent( u_component(y,self.un2), -self.u2, TOL )
        equivalent( u_component(mag_y,self.un2), self.u2, TOL )

        # Intermediate tests (these use the i_component value)       
        y = +self.un_12
        mag_y = magnitude(y)
        equivalent( value(mag_y), self.x_12, TOL )
        equivalent( u_component(y,self.un_12), self.u_12, TOL )
        equivalent( u_component(mag_y,self.un_12), self.u_12, TOL )
        
        y = -1 * self.un_12
        mag_y = magnitude(y)
        equivalent( value(mag_y), self.x_12, TOL )
        equivalent( u_component(y,self.un_12), -self.u_12, TOL )
        equivalent( u_component(mag_y,self.un_12), self.u_12, TOL )


#-----------------------------------------------------
class TrigTestsReal(unittest.TestCase):

    def setUp(self):
        self.x = 0.7
        self.u = 0.3
        self.un = ureal(self.x,self.u)

        self.x1 = -0.2
        self.u1 = 0.1
        self.un1 = ureal(self.x1,self.u1)

        # Using `result` means that the i_components
        # buffer is used for the intermediate components.
        self.x2 = self.x * self.x1
        self.un2 = result( self.un * self.un1 )
        
    def testSine(self):
        # There are no special cases of inputs to test
        y = sin(self.un)
        v = value(y)
        u = uncertainty(y)
        df = dof(y)
        equivalent( v, math.sin( self.x ), TOL )
        equivalent( u, math.cos( self.x ) * self.u, TOL )
        self.assertTrue( math.isinf(df) )

        # This will test an intermediate component of uncertainty
        vv = math.sin(self.x2)
        uu = math.cos(self.x2) * uncertainty( self.un2 )
        y = sin( self.un2 )
        v = value(y)
        equivalent( v,  vv, TOL  )
        equivalent( uu, uncertainty(y), TOL )
        equivalent( uu, u_component(y,self.un2), TOL )
        
    def testCosine(self):
        # There are no special cases of inputs to test
        y = cos(self.un)
        v = value(y)
        u = uncertainty(y)
        df = dof(y)
        equivalent( v, math.cos( self.x ),TOL )
        equivalent( u, math.sin( self.x) * self.u, TOL )
        self.assertTrue( math.isinf(df) )
        
        # This will test an intermediate component of uncertainty
        vv = math.cos(self.x2)
        uu = -math.sin(self.x2) * uncertainty( self.un2 )
        y = cos( self.un2 )
        v = value(y)
        equivalent( v,  vv, TOL  )
        equivalent( uu, uncertainty(y), TOL )
        equivalent( uu, u_component(y,self.un2), TOL )

    def testTangent(self):
        # There are no special cases of inputs to test
        y = tan(self.un)
        v = value(y)
        u = uncertainty(y)
        df = dof(y)
        equivalent( v, math.tan( self.x ), TOL )
        uu = self.u/(math.cos( self.x )**2)
        equivalent( u, uu, TOL )
        self.assertTrue( math.isinf(df) )

        # This will test an intermediate component of uncertainty
        vv = math.tan(self.x2)
        uu = uncertainty( self.un2 ) /(math.cos( self.x2 )**2)
        y = tan( self.un2 )
        v = value(y)
        equivalent( v,  vv, TOL  )
        equivalent( uu, uncertainty(y), TOL )
        equivalent( uu, u_component(y,self.un2), TOL )
        
class FunctionTestsReal(unittest.TestCase):
    def setUp(self):
        self.x1 = 1.7
        self.u1 = 0.1
        self.un1 = ureal(self.x1,self.u1) 
        
        self.x2 = 0.5
        self.u2 = 0.4
        self.un2 = ureal(self.x2,self.u2)
        
        self.x3 = -1.0
        self.u3 = 0.4
        self.un3 = ureal(self.x3,self.u3)

        self.x4 = 0.0
        self.u4 = 0.2
        self.un4 = ureal(self.x4,self.u4)
        
        self.x5 = 1.0
        self.u5 = 0.2
        self.un5 = ureal(self.x5,self.u5)

        # Create intermediate values for u_component checking
        self.x_12 = self.x1 * self.x2
        self.u_12 = math.sqrt((self.x1 * self.u2)**2 + (self.x2 * self.u1)**2)
        
        # By using `result` here the i_components buffer will be used
        # in the intermediate tests. 
        self.un_12 = result( self.un1 * self.un2 ) 
        
        self.un_22 = result( self.un2 / ureal(.3,.1) )

    def testLog(self):
        y = log(self.un1)
        v = value(y)
        u = uncertainty(y)
        equivalent( v, math.log( self.x1 ), TOL )
        equivalent( u, self.u1 / self.x1, TOL )

        y = log(self.un2)
        v = value(y)
        u = uncertainty(y)
        equivalent( v, math.log( self.x2 ), TOL )
        equivalent( u, self.u2 / self.x2, TOL )

        self.assertRaises(ValueError,log,self.un3)

        # This will test an intermediate component of uncertainty
        vv = math.log(self.x_12)
        uu = 1/value(self.x_12) * uncertainty( self.un_12 )
        y = log( self.un_12 )
        v = value(y)
        equivalent( v,  vv, TOL  )
        equivalent( uu, uncertainty(y), TOL )
        equivalent( uu, u_component(y,self.un_12), TOL )

    def testLog10(self):
        y = log10(self.un1)
        v = value(y)
        u = uncertainty(y)
        equivalent( v, LOG10_E * math.log( self.x1 ), TOL )
        equivalent( u, LOG10_E * self.u1 / self.x1, TOL )

        y = log10(self.un2)
        v = value(y)
        u = uncertainty(y)
        equivalent( v, LOG10_E * math.log( self.x2 ), TOL )
        equivalent( u, LOG10_E * self.u2 / self.x2, TOL )

        self.assertRaises(ValueError,log10,self.un3)

        # This will test an intermediate component of uncertainty
        vv = math.log(self.x_12)
        uu = 1/value(self.x_12) * uncertainty( self.un_12 )
        y = log10( self.un_12 )
        v = value(y)
        equivalent( v,  LOG10_E * vv, TOL  )
        equivalent( LOG10_E * uu, uncertainty(y), TOL )
        equivalent( LOG10_E * uu, u_component(y,self.un_12), TOL )
        
    def testSqrt(self):
        y = sqrt(self.un1)
        v = value(y)
        vv = math.sqrt(self.x1)
        uu = self.u1 / ( 2.0 * vv )
        
        equivalent( v,  vv, TOL  )
        equivalent( uu, u_component(y,self.un1), TOL )

        y = sqrt(self.un2) 
        v = value(y)
        vv = math.sqrt(self.x2)
        uu = self.u2 / ( 2.0 * vv )
        
        equivalent( v,  vv, TOL  )
        equivalent( uu, u_component(y,self.un2), TOL )

        self.assertRaises(ValueError,sqrt,self.un3)

        # This will test an intermediate component of uncertainty
        vv = math.sqrt(self.x_12)
        uu = 0.5/vv * uncertainty( self.un_12 )
        y = sqrt( self.un_12 )
        v = value(y)
        equivalent( v,  vv, TOL  )
        equivalent( uu, uncertainty(y), TOL )
        equivalent( uu, u_component(y,self.un_12), TOL )

    def testExp(self):
        # There are no special cases of inputs to test
        vv = math.exp(self.x1)
        uu = self.u1 * vv
        y = exp( self.un1 )
        v = value(y)
        equivalent( v,  vv, TOL  )
        equivalent( uu, u_component(y,self.un1), TOL )
        equivalent( uu, uncertainty(y), TOL )

        vv = math.exp(self.x2)
        uu = self.u2 * vv
        y = exp( self.un2 )
        v = value(y)
        equivalent( v,  vv, TOL  )
        equivalent( uu, u_component(y,self.un2), TOL )

        vv = math.exp(self.x3)
        uu = self.u3 * vv
        y = exp( self.un3 )
        v = value(y)
        equivalent( v,  vv, TOL  )
        equivalent( uu, u_component(y,self.un3), TOL )

        # This will test an intermediate component of uncertainty
        vv = math.exp(self.x_12)
        uu = vv * uncertainty( self.un_12 )
        y = exp( self.un_12 )
        v = value(y)
        equivalent( v,  vv, TOL  )
        equivalent( uu, uncertainty(y), TOL )
        equivalent( uu, u_component(y,self.un_12), TOL )
        
    def testMagSquared(self):
        # when argument is ureal, it returns self*self
        v4 = 2
        u4 = 0.5
        z4 = ureal(v4,u4)
        z4_mag = mag_squared(z4)
        self.assertTrue(
            equivalent(value(z4_mag),v4**2)
        )
        self.assertTrue(
            equivalent(uncertainty(z4_mag),2*v4*u4)
        )

    def testMagnitude(self):
        v1 = 1.5
        u1 = 1
        v2 = -.5
        u2 = 1
        v3 = 0.0
        u3 = 1
        z1 = ureal(v1,u1)
        z1_mag = magnitude(z1)
        self.assertTrue(
            equivalent(value(z1_mag),v1)
        )
        self.assertTrue(
            equivalent(uncertainty(z1),u1)
        )
        
        z = ureal(v2,u2)
        z_mag = magnitude(z)
        self.assertTrue(
            equivalent(value(z_mag),abs(v2) )
        )
        self.assertTrue(
            equivalent(uncertainty(z),u2)
        )
        z = ureal(v3,u3)
        self.assertRaises( ZeroDivisionError, magnitude, z )

    def testPhase(self):
        v1 = 1.5
        u1 = 1
        z1 = ureal(v1,u1)
        self.assertEqual( phase(z1), 0.0 )
        
#-----------------------------------------------------
class TestGetCovariance(unittest.TestCase):

    def test_with_mixed_unumbers(self):
    
        ux1 = 2
        x1 = ureal(1,ux1,independent=False)
        uz1 = .5
        z1 = ureal(1,uz1,independent=False)
        
        r1 = .1
        set_correlation(r1,z1,x1)
        check_r = get_covariance(z1,x1)
        self.assertTrue( equivalent(r1,check_r) )
        
        r2 = -.3
        set_correlation(r2,z1,x1)
        check_r = get_covariance(z1,x1)
        self.assertTrue( equivalent(r2,check_r) )

    def test_simple_correlation(self):
        x1 = ureal(0,1,5,None,independent=True)
        x2 = ureal(0,1,5,None,independent=True)

        self.assertRaises(
            RuntimeError,
            set_correlation_real,x1,x2,.1
        )

        # self correlation must be unity
        self.assertRaises(
            RuntimeError,
            set_correlation_real,x1,x1,.1
        )

        x1 = ureal(0,1,5,None,independent=False)
        x2 = ureal(0,1,5,None,independent=False)
        
        # Self correlation
        self.assertEqual( 1, get_correlation_real(x1,x1) )
        
        r = 0.65
        set_correlation_real(x1,x2,r)
        
        self.assertTrue( equivalent( r, get_correlation_real(x1,x2) ) )

        # When we delete one node the other remains
        x1_uid = x1._node.uid
        x2_uid = x2._node.uid
        self.assertTrue( x1_uid in context._context._registered_leaf_nodes )
        self.assertTrue( x2_uid in context._context._registered_leaf_nodes )
        del x1
        self.assertTrue( x1_uid not in context._context._registered_leaf_nodes )
        self.assertTrue( x2_uid in context._context._registered_leaf_nodes )

        # self.assertEqual(1, len(c._registered_leaf_nodes) )
        # self.assertTrue( x2._node.uid in c._registered_leaf_nodes )
        
        # Correlation matrix is not trimmed when `x1` is destroyed
        self.assertEqual(2, len(x2._node.correlation) )
        self.assertTrue( x2_uid in x2._node.correlation )
        self.assertTrue( x1_uid in x2._node.correlation )
        
        self.assertEqual(2, len(x2._node.correlation) )
        self.assertEqual(1.0, x2._node.correlation[x2_uid] )
        
#-----------------------------------------------------
class ArcTrigTests(unittest.TestCase):
    def setUp(self):
        self.x = 0.7
        self.u = 0.3
        self.un = ureal(self.x,self.u)
        
    def test(self):
        PI = math.pi
        arg1 = 0.75             # 1st quadrant
        arg2 = arg1 + PI / 2    # 2nd quadrant
        arg3 = arg1-PI          # 3rd quadrant
        arg4 = arg1 - PI/2      # 4th quadrant 
        u = 0.05

        x1 = ureal(arg1,u)        
        x2 = ureal(arg2,u)        
        x3 = ureal(arg3,u)        
        x4 = ureal(arg4,u)

        # atan2 test ---------------------
        # u = y/x
        df_dy = lambda y,x: x/(x**2 + y**2)
        df_dx = lambda y,x: -y/(x**2 + y**2)
        
        x_cpt = result( cos(x1) )
        y_cpt = result( sin(x1) )
        x = atan2(y_cpt,x_cpt)
        
        cos_x1 = math.cos(arg1)
        sin_x1 = math.sin(arg1)
        
        equivalent(value(x),arg1,TOL)
        equivalent(u_component(x,x1),u,TOL)

        # mixed argument types
        x = atan2(y_cpt,cos_x1)
        equivalent(value(x),arg1,TOL)
        u_y = df_dy(sin_x1,cos_x1) * uncertainty(y_cpt)
        equivalent(u_component(x,y_cpt),u_y,TOL)

        x = atan2(sin_x1,x_cpt)
        equivalent(value(x),arg1,TOL)
        u_x = df_dx(sin_x1,cos_x1) * uncertainty(x_cpt)
        equivalent(u_component(x,x_cpt),u_x,TOL)
        
        # intermediate -----
        x = atan2(y_cpt,x_cpt)
        xv = value(x_cpt)
        yv = value(y_cpt)
        den = xv**2 + yv**2
        equivalent(
            u_component(x,x_cpt),
            uncertainty(x_cpt)* -yv/den,
            TOL)
        equivalent(
            u_component(x,y_cpt),
            uncertainty(y_cpt)* xv/den,
            TOL)

        # illegal 
        self.assertRaises(TypeError,atan2,y_cpt,1+5j) 
        self.assertRaises(TypeError,atan2,1+5j,x_cpt,) 
        
        # trivial 
        y = atan2(ureal(0,1),ureal(0,1))
        self.assertEqual( value(y), 0 )
        self.assertEqual( uncertainty(y), 0 )
        self.assertEqual( dof(y), inf )   
 
        y = atan2(0,ureal(0,1))
        self.assertEqual( value(y), math.atan2(0,0) )
        self.assertEqual( uncertainty(y), 0 )
        self.assertEqual( dof(y), inf )   
 
        y = atan2(ureal(0,1),0)
        self.assertEqual( value(y), math.atan2(0,0) )
        self.assertEqual( uncertainty(y), 0 )
        self.assertEqual( dof(y), inf ) 
        
        # ----------------------        
        x_cpt = result( cos(x2) )
        y_cpt = result( sin(x2) )
        x = atan2(y_cpt,x_cpt)
        
        cos_x = math.cos(arg2)
        sin_x = math.sin(arg2)
        equivalent(value(x),arg2,TOL)
        equivalent(u_component(x,x2),u,TOL)
        
        # intermediate -----
        xv = value(x_cpt)
        yv = value(y_cpt)
        den = xv**2 + yv**2
        equivalent(
            u_component(x,x_cpt),
            uncertainty(x_cpt)* -yv/den,
            TOL)
        equivalent(
            u_component(x,y_cpt),
            uncertainty(y_cpt)* xv/den,
            TOL)

        # mixed argument types
        x = atan2(y_cpt,cos_x)
        equivalent(value(x),arg2,TOL)
        u_y = df_dy(sin_x,cos_x) * uncertainty(y_cpt)
        equivalent(u_component(x,y_cpt),u_y,TOL)

        x = atan2(sin_x,x_cpt)
        equivalent(value(x),arg2,TOL)
        u_x = df_dx(sin_x,cos_x) * uncertainty(x_cpt)
        equivalent(u_component(x,x_cpt),u_x,TOL)
        
        # ----------------------        
        x_cpt = result( cos(x3) )
        y_cpt = result( sin(x3) )
        x = atan2(y_cpt,x_cpt)
        
        cos_x = math.cos(arg3)
        sin_x = math.sin(arg3)
        equivalent(value(x),arg3,TOL)
        equivalent(u_component(x,x3),u,TOL)

        # intermediate -----
        xv = value(x_cpt)
        yv = value(y_cpt)
        den = xv**2 + yv**2
        equivalent(
            u_component(x,x_cpt),
            uncertainty(x_cpt)* -yv/den,
            TOL)
        equivalent(
            u_component(x,y_cpt),
            uncertainty(y_cpt)* xv/den,
            TOL)

        # mixed argument types
        x = atan2(y_cpt,cos_x)
        equivalent(value(x),arg3,TOL)
        u_y = df_dy(sin_x,cos_x) * uncertainty(y_cpt)
        equivalent(u_component(x,y_cpt),u_y,TOL)

        x = atan2(sin_x,x_cpt)
        equivalent(value(x),arg3,TOL)
        u_x = df_dx(sin_x,cos_x) * uncertainty(x_cpt)
        equivalent(u_component(x,x_cpt),u_x,TOL)
        
        # ----------------------        
        x_cpt = result( cos(x4) )
        y_cpt = result( sin(x4) )
        x = atan2(y_cpt,x_cpt)
        
        cos_x = math.cos(arg4)
        sin_x = math.sin(arg4)
        equivalent(value(x),arg4,TOL)
        equivalent(u_component(x,x4),u,TOL)

        # intermediate -----
        xv = value(x_cpt)
        yv = value(y_cpt)
        den = xv**2 + yv**2
        equivalent(
            u_component(x,x_cpt),
            uncertainty(x_cpt)* -yv/den,
            TOL)
        equivalent(
            u_component(x,y_cpt),
            uncertainty(y_cpt)* xv/den,
            TOL)

        # mixed argument types
        x = atan2(y_cpt,cos_x)
        equivalent(value(x),arg4,TOL)
        u_y = df_dy(sin_x,cos_x) * uncertainty(y_cpt)
        equivalent(u_component(x,y_cpt),u_y,TOL)

        x = atan2(sin_x,x_cpt)
        equivalent(value(x),arg4,TOL)
        u_x = df_dx(sin_x,cos_x) * uncertainty(x_cpt)
        equivalent(u_component(x,x_cpt),u_x,TOL)
        
        # asin test ---------------------------------
        y_cpt = result( sin(x1) )
        x = asin(y_cpt)
        equivalent(value(x),arg1,TOL)
        equivalent(uncertainty(x),u,TOL)
        equivalent(u_component(x,x1),u,TOL)
        
        # intermediate -----
        equivalent(
            u_component(x,y_cpt),
            uncertainty(y_cpt)/math.sqrt(1-value(y_cpt)**2),
            TOL)
        

        # Maps to PI/2 - arg1        
        y_cpt = result( sin(x2) )
        x = asin(y_cpt)
        equivalent(value(x),PI/2-arg1,TOL)
        equivalent(u_component(x,x2),-u,TOL)
        equivalent(uncertainty(x),u,TOL)
        
        # intermediate -----
        equivalent(
            u_component(x,y_cpt),
            uncertainty(y_cpt)/math.sqrt(1-value(y_cpt)**2),
            TOL)

        # Maps to  - arg1        
        y_cpt = result( sin(x3) )
        x = asin(y_cpt)
        equivalent(value(x),-arg1,TOL)
        equivalent(u_component(x,x3),-u,TOL)
        equivalent(uncertainty(x),u,TOL)
        
        # intermediate -----
        equivalent(
            u_component(x,y_cpt),
            uncertainty(y_cpt)/math.sqrt(1-value(y_cpt)**2),
            TOL)

        y_cpt = result( sin(x4) )
        x = asin(y_cpt)
        equivalent(value(x),arg4,TOL)
        equivalent(u_component(x,x4),u,TOL)
        equivalent(uncertainty(x),u,TOL)
        
        # intermediate -----
        equivalent(
            u_component(x,y_cpt),
            uncertainty(y_cpt)/math.sqrt(1-value(y_cpt)**2),
            TOL)

        # acos test ---------------------------------
        y_cpt = result( cos(x1) )
        x = acos(y_cpt)
        equivalent(value(x),arg1,TOL)
        equivalent(u_component(x,x1),u,TOL)
        equivalent(uncertainty(x),u,TOL)

        # intermediate -----
        equivalent(
            u_component(x,y_cpt),
            -uncertainty(y_cpt)/math.sqrt(1-value(y_cpt)**2),
            TOL)

        y_cpt = result( cos(x2) )
        x = acos(y_cpt)
        equivalent(value(x),arg2,TOL)
        equivalent(u_component(x,x2),u,TOL)
        equivalent(uncertainty(x),u,TOL)
        
        # intermediate -----
        equivalent(
            u_component(x,y_cpt),
            -uncertainty(y_cpt)/math.sqrt(1-value(y_cpt)**2),
            TOL)

        # Maps to  PI - arg1        
        y_cpt = result( cos(x3) )
        x = acos(y_cpt)
        equivalent(value(x),PI - arg1,TOL)
        equivalent(u_component(x,x3),-u,TOL)
        equivalent(uncertainty(x),u,TOL)
        
        # intermediate -----
        equivalent(
            u_component(x,y_cpt),
            -uncertainty(y_cpt)/math.sqrt(1-value(y_cpt)**2),
            TOL)

        # Maps to  PI/2 - arg1        
        y_cpt = result( cos(x4) )
        x = acos(y_cpt)
        equivalent(value(x),PI/2 - arg1,TOL)
        equivalent(u_component(x,x4),-u,TOL)
        equivalent(uncertainty(x),u,TOL)
        
        # intermediate -----
        equivalent(
            u_component(x,y_cpt),
            -uncertainty(y_cpt)/math.sqrt(1-value(y_cpt)**2),
            TOL)

        # atan test ---------------------------------
        t = result( tan(x1) )
        x = atan(t)
        equivalent(value(x),arg1,TOL)
        equivalent(u_component(x,x1),u,TOL)
        equivalent(uncertainty(x),u,TOL)

        # intermediate -----
        equivalent(
            u_component(x,t),
            uncertainty(t)/(1+value(t)**2),
            TOL)

        # mapped to arg1 - PI/2
        t = result( tan(x2) )
        x = atan(t)
        equivalent(value(x),arg1 - PI/2,TOL)
        equivalent(u_component(x,x2),u,TOL)
        equivalent(uncertainty(x),u,TOL)

        # intermediate -----
        equivalent(
            u_component(x,t),
            uncertainty(t)/(1+value(t)**2),
            TOL)

        # mapped to arg1
        t = result( tan(x3) )
        x = atan(t)
        equivalent(value(x),arg1,TOL)
        equivalent(u_component(x,x3),u,TOL)
        equivalent(uncertainty(x),u,TOL)

        # intermediate -----
        equivalent(
            u_component(x,t),
            uncertainty(t)/(1+value(t)**2),
            TOL)

        t = result( tan(x4) )
        x = atan(t)
        equivalent(value(x),arg4,TOL)
        equivalent(u_component(x,x4),u,TOL)
        equivalent(uncertainty(x),u,TOL)

        # intermediate -----
        equivalent(
            u_component(x,t),
            uncertainty(t)/(1+value(t)**2),
            TOL)

        # illegal input values -----------------------
        illegal1 = ureal(1.01,1)
        illegal2 = ureal(-1.01,1)

        self.assertRaises(ValueError,asin,illegal1)        
        self.assertRaises(ValueError,asin,illegal2)        
        self.assertRaises(ValueError,acos,illegal1)        
        self.assertRaises(ValueError,acos,illegal2)        

#-----------------------------------------------------
class HyperbolicTrigTestsReal(unittest.TestCase):
    def setUp(self):
        self.x = 0.44
        self.u = 0.13
        self.un = ureal(self.x,self.u)

        self.x1 = -0.43
        self.u1 = 0.2
        self.un1 = ureal(self.x1,self.u1)

        self.x2 = self.x * self.x1
        self.un2 = result( self.un * self.un1 )
        
    def testhSine(self):
        y = sinh(self.un)
        v = value(y)
        u = uncertainty(y)
        df = dof(y)
        equivalent( v, math.sinh( self.x ), TOL )
        equivalent( u, math.cosh( self.x ) * self.u, TOL )
        self.assertTrue( math.isinf(df) )

        # This will test an intermediate component of uncertainty
        vv = math.sinh(self.x2)
        uu = math.cosh(self.x2) * uncertainty( self.un2 )
        y = sinh( self.un2 )
        v = value(y)
        equivalent( v,  vv, TOL  )
        equivalent( uu, uncertainty(y), TOL )
        equivalent( uu, u_component(y,self.un2), TOL )
        
    def testhCosine(self):
        y = cosh(self.un)
        v = value(y)
        u = uncertainty(y)
        df = dof(y)
        equivalent( v, math.cosh( self.x ),TOL )
        equivalent( u, math.sinh( self.x) * self.u, TOL )
        self.assertTrue( math.isinf(df) )
        
        # This will test an intermediate component of uncertainty
        vv = math.cosh(self.x2)
        uu = math.sinh(self.x2) * uncertainty( self.un2 )
        y = cosh( self.un2 )
        v = value(y)
        equivalent( v,  vv, TOL  )
        equivalent( abs( uu ), uncertainty(y), TOL )
        equivalent( uu, u_component(y,self.un2), TOL )

    def testhTangent(self):
        y = tanh(self.un)
        v = value(y)
        u = uncertainty(y)
        df = dof(y)
        equivalent( v, math.tanh( self.x ), TOL )
        uu = self.u/(math.cosh( self.x )**2)
        equivalent( u, uu, TOL )
        self.assertTrue( math.isinf(df) )

        # This will test an intermediate component of uncertainty
        vv = math.tanh(self.x2)
        uu = uncertainty( self.un2 ) /(math.cosh( self.x2 )**2)
        y = tanh( self.un2 )
        v = value(y)
        equivalent( v,  vv, TOL  )
        equivalent( uu, uncertainty(y), TOL )
        equivalent( uu, u_component(y,self.un2), TOL )

#-----------------------------------------------------
class InverseHyperbolicTrigTestsReal(unittest.TestCase):
    def setUp(self):
        self.x = 1.44
        self.u = 0.13
        self.un = ureal(self.x,self.u)

        self.x1 = 2.03
        self.u1 = 0.2
        self.un1 = ureal(self.x1,self.u1)

        self.x2 = self.x * self.x1
        self.un2 = result( self.un * self.un1 )

        self.x3 = 0.44
        self.u3 = 0.23
        self.un3 = ureal(self.x3,self.u3)

        self.x4 = .13
        self.u4 = 0.2
        self.un4 = ureal(self.x4,self.u4)

        self.x5 = self.x3 * self.x4
        self.un5 = result( self.un3 * self.un4 )
        
    def testArcSineh(self):
        log_form = lambda x: math.log(x + math.sqrt(1 + x**2))
        derivative = lambda x: 1./math.sqrt(1 + x**2)
        
        y = asinh(self.un)
        v = value(y)
        u = uncertainty(y)
        df = dof(y)
        equivalent( v, log_form( self.x ), TOL )
        equivalent( u, derivative( self.x ) * self.u, TOL )
        self.assertTrue( math.isinf(df) )

        # This will test an intermediate component of uncertainty
        vv = log_form(self.x2)
        uu = derivative(self.x2) * uncertainty( self.un2 )
        y = asinh( self.un2 )
        v = value(y)
        equivalent( v,  vv, TOL  )
        equivalent( uu, uncertainty(y), TOL )
        equivalent( uu, u_component(y,self.un2), TOL )
        
    def testArcCosineh(self):
        log_form = lambda x: math.log(x + math.sqrt(x**2 - 1))
        derivative = lambda x: 1./(math.sqrt(x-1) * math.sqrt(x+1))

        y = acosh(self.un)
        v = value(y)
        u = uncertainty(y)
        df = dof(y)
        equivalent( v, log_form( self.x ),TOL )
        equivalent( u, derivative( self.x) * self.u, TOL )
        self.assertTrue( math.isinf(df) )
        
        # This will test an intermediate component of uncertainty
        vv = log_form(self.x2)
        uu = derivative(self.x2) * uncertainty( self.un2 )
        y = acosh( self.un2 )
        v = value(y)
        equivalent( v,  vv, TOL  )
        equivalent( abs( uu ), uncertainty(y), TOL )
        equivalent( uu, u_component(y,self.un2), TOL )

    def testArcTangenth(self):
        log_form = lambda x: math.log((1+x)/(1-x))/2.
        derivative = lambda x: 1./(1 - x**2)

        y = atanh(self.un3)
        v = value(y)
        u = uncertainty(y)
        df = dof(y)
        equivalent( v, log_form( self.x3 ), TOL )
        uu = derivative(self.x3) * self.u3
        equivalent( u, uu, TOL )
        self.assertTrue( math.isinf(df) )

        # This will test an intermediate component of uncertainty
        vv = log_form(self.x5)
        uu = uncertainty( self.un5 ) * derivative( self.x5 )
        y = atanh( self.un5 )
        v = value(y)
        equivalent( v,  vv, TOL  )
        equivalent( uu, uncertainty(y), TOL )
        equivalent( uu, u_component(y,self.un5), TOL )
        
#-----------------------------------------------------
class GuideExampleH2(unittest.TestCase):

    def test_no_dof(self):
        TOL = 1E-10

        v = ureal(4.999,0.0032,independent=False)
        i = ureal(0.019661,0.0000095,independent=False)
        phi = ureal(1.04446,0.00075,independent=False)

        set_correlation(-0.36,v,i)
        set_correlation(0.86,v,phi)
        set_correlation(-0.65,i,phi)
        
        r = v * cos(phi)/ i
        x = v * sin(phi)/ i
        z = v / i

        equivalent( uncertainty(r),0.0699787279884,TOL)
        equivalent( uncertainty(x),0.295716826846,TOL)
        equivalent( uncertainty(z),0.236602971835,TOL)

        equivalent( math.sqrt(welch_satterthwaite(r)[0]),uncertainty(r),TOL)
        equivalent( math.sqrt(welch_satterthwaite(x)[0]),uncertainty(x),TOL)
        equivalent( math.sqrt(welch_satterthwaite(z)[0]),uncertainty(z),TOL)

        equivalent( get_correlation(r,x),-0.591484610819,TOL)
        equivalent( get_correlation(x,z),0.992797472722,TOL)
        equivalent( get_correlation(r,z),-0.490623905441,TOL)

    def test_with_dof(self):
        TOL = 1E-10

        # It is illegal to do this now
        # v = ureal(4.999,0.0032,5,independent=False)
        # i = ureal(0.019661,0.0000095,5,independent=False)
        # phi = ureal(1.04446,0.00075,5,independent=False)
        # Instead, we must use multiple_ureal
        
        # Once the inputs are in an ensemble we can calculate dof
        x = [4.999,0.019661,1.04446]
        u = [0.0032,0.0000095,0.00075]
        v,i,phi = multiple_ureal(x,u,5)
        
        set_correlation(-0.36,v,i)
        set_correlation(0.86,v,phi)
        set_correlation(-0.65,i,phi)

        r = v * cos(phi)/ i
        x = v * sin(phi)/ i
        z = v / i

        # The presence of finite DoF should not alter previous results
        equivalent( uncertainty(r),0.0699787279884,TOL)
        equivalent( uncertainty(x),0.295716826846,TOL)
        equivalent( uncertainty(z),0.236602971835,TOL)

        equivalent( math.sqrt(welch_satterthwaite(r)[0]),math.sqrt( std_variance_real(r) ),TOL)
        equivalent( math.sqrt(welch_satterthwaite(x)[0]),uncertainty(x),TOL)
        equivalent( math.sqrt(welch_satterthwaite(z)[0]),uncertainty(z),TOL)

        equivalent( get_correlation(r,x),-0.591484610819,TOL)
        equivalent( get_correlation(x,z),0.992797472722,TOL)
        equivalent( get_correlation(r,z),-0.490623905441,TOL)

        equivalent( dof(r),5,TOL)
        equivalent( dof(x),5,TOL)
        equivalent( dof(z),5,TOL)
        
        # Add another influence and it breaks again, but earlier results OK
        f = ureal(1,0.0032,5,independent=False)
        self.assertRaises(RuntimeError,set_correlation,0.3,f,i)
        
        equivalent( dof(r),5,TOL)
        equivalent( dof(x),5,TOL)
        equivalent( dof(z),5,TOL)

    def test_with_dof_2(self):
        """
        Same test as above, but now ensure that the 
        ensemble calculation can deal with influences 
        that are not in sequence
        """
        TOL = 1E-10

        # The dummy variables mean that those included 
        # in the ensemble will not be a consecutive 
        # series of IDs, provided we include the 
        # dummies in the equations, as is done below
        # with zero weight.
        v = ureal(4.999,0.0032,5,independent=False)
        dummy_1 = ureal(1,1)
        i = ureal(0.019661,0.0000095,5,independent=False)
        dummy_2 = ureal(1,1)
        phi = ureal(1.04446,0.00075,5,independent=False)
        dummy_3 = ureal(1,1)

        real_ensemble( [v,i,phi], 5 )

        set_correlation(-0.36,v,i)
        set_correlation(0.86,v,phi)
        set_correlation(-0.65,i,phi)

        r = v * cos(phi)/ i + (0*dummy_1)
        x = v * sin(phi)/ i + (0*dummy_2)
        z = v / i + (0*dummy_3)

        # The presence of finite DoF should not alter previous results
        equivalent( uncertainty(r),0.0699787279884,TOL)
        equivalent( uncertainty(x),0.295716826846,TOL)
        equivalent( uncertainty(z),0.236602971835,TOL)

        equivalent( math.sqrt(welch_satterthwaite(r)[0]),math.sqrt( std_variance_real(r) ),TOL)
        equivalent( math.sqrt(welch_satterthwaite(x)[0]),uncertainty(x),TOL)
        equivalent( math.sqrt(welch_satterthwaite(z)[0]),uncertainty(z),TOL)

        equivalent( get_correlation(r,x),-0.591484610819,TOL)
        equivalent( get_correlation(x,z),0.992797472722,TOL)
        equivalent( get_correlation(r,z),-0.490623905441,TOL)

        equivalent( dof(r),5,TOL)
        equivalent( dof(x),5,TOL)
        equivalent( dof(z),5,TOL) 
        
    def test_with_dof_multiple_ureal(self):
        TOL = 1E-10

        x = [4.999,0.019661,1.04446]
        u = [0.0032,0.0000095,0.00075]
        labels = ['v','i','phi']

        v,i,phi = multiple_ureal(x,u,5,labels)
        
        set_correlation(-0.36,v,i)
        set_correlation(0.86,v,phi)
        set_correlation(-0.65,i,phi)

        r = v * cos(phi)/ i
        x = v * sin(phi)/ i
        z = v / i

        # The presence of finite DoF should not alter previous results
        equivalent( uncertainty(r),0.0699787279884,TOL)
        equivalent( uncertainty(x),0.295716826846,TOL)
        equivalent( uncertainty(z),0.236602971835,TOL)

        equivalent( math.sqrt(welch_satterthwaite(r)[0]),math.sqrt( std_variance_real(r) ),TOL)
        equivalent( math.sqrt(welch_satterthwaite(x)[0]),uncertainty(x),TOL)
        equivalent( math.sqrt(welch_satterthwaite(z)[0]),uncertainty(z),TOL)

        equivalent( get_correlation(r,x),-0.591484610819,TOL)
        equivalent( get_correlation(x,z),0.992797472722,TOL)
        equivalent( get_correlation(r,z),-0.490623905441,TOL)

        # Dof calculation should be legal
        equivalent( dof(r),5,TOL)
        equivalent( dof(x),5,TOL)
        equivalent( dof(z),5,TOL)
        
        # Add another influence and it breaks, but earlier results OK
        f = ureal(1,0.0032,5,independent=False)
        self.assertRaises(RuntimeError,set_correlation,0.3,f,i)
                
        equivalent( dof(r),5,TOL)
        equivalent( dof(x),5,TOL)
        equivalent( dof(z),5,TOL)
        

#-----------------------------------------------------
class TestMultipleUNsWithRealConstants(unittest.TestCase):
    """
    When defining ensembles of real or complex uncertain numbers,
    some elements might in fact have no uncertainty! In which
    case they are defined as constants.
    
    This seems a bit artificial: in effect the user is
    passing an uncertain number as part of an ensemble
    when it is not! However, we may have this situation
    in fitting scenarios (perhaps a parameter comes
    back with no uncertainty due to some odd data).
    So it is provided for. The creation of a constant
    should not harm subsequent calculations.
    
    """
    def test_ureal(self):
    
        x = [4.999,19.661E-3,1.04446]
        u = [3.2E-3,0,7.5E-4]
        labels = ['V','I','phi']
        
        v,i,phi = multiple_ureal(x,u,4,labels)
        
        self.assertTrue(v.x == x[0])
        self.assertTrue(v.u == u[0])
        self.assertTrue(v.df == 4)

        self.assertTrue(i.x == x[1])
        self.assertTrue(i.u == u[1])
        self.assertTrue(i.df == inf)
        
        self.assertTrue(phi.x == x[2])
        self.assertTrue(phi.u == u[2])
        self.assertTrue(phi.df == 4)

        # We can set a zero correlation ...
        set_correlation(0.0,v,i)
        set_correlation(0.0,i,phi)
        set_correlation(0.0,phi,v)
        
        # but not a finite one to a constant
        self.assertRaises(RuntimeError,set_correlation,0.1,v,i)
        self.assertRaises(RuntimeError,set_correlation,0.1,phi,i)
        set_correlation(0.1,phi,v)


#-----------------------------------------------------
class TestRealEnsemble(unittest.TestCase):

    def test_append_real_ensemble(self):

        df = 4
        x1 = ureal(0,1,df,independent=False)
        x2 = ureal(0,1,df,independent=False)
        
        real_ensemble([x1,x2],df=df)
        self.assertEqual(x1._node.ensemble,x2._node.ensemble)
        
        x3 = ureal(0,1,df,independent=False)
        self.assertNotEqual(x1._node.ensemble,x3._node.ensemble)
        append_real_ensemble(x1,x3)
        self.assertEqual(x1._node.ensemble,x3._node.ensemble)
        
        _ensemble = x1._node.ensemble 
        
        self.assertTrue( x3._node.uid in _ensemble )
        self.assertTrue( x2._node.uid in _ensemble )
        self.assertTrue( x1._node.uid in _ensemble )
        
        self.assertEqual(x1._node.ensemble,x2._node.ensemble)
        self.assertEqual(x1._node.ensemble,x3._node.ensemble)

        del x3
        
        self.assertEqual(x1._node.ensemble,x2._node.ensemble)
        
    def test_illegal_real_ensemble(self):
    
        # influences with different DoF 
        x1 = ureal(0,1,3,independent=False)
        x2 = ureal(0,1,4,independent=False)
                
        self.assertRaises(
            AssertionError,
            real_ensemble,
            [x1,x2], 
            4
        )
        
        x3 = ureal(0,1,4,independent=False)
        real_ensemble([x2,x3],4)
        self.assertRaises(
            AssertionError,
            append_real_ensemble,
            x2,x1
        )
        
        # non-elementary influences 
        y = x1+ureal(0,1,3,independent=False) 
        self.assertRaises(
            AssertionError,
            real_ensemble,
            [x1,y], 3
        )
  
        self.assertRaises(
            AssertionError,
            append_real_ensemble,
            x1,y
        )
#============================================================================
if(__name__== '__main__'):

    unittest.main()    # Runs all test methods starting with 'test'